"""
Select or remove data by some criteria
"""

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Callable, Literal, Optional, Union
from uuid import uuid4

import numpy as np
import pandas as pd
import polars as pl

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame, PolarsDataFrame, SparkDataFrame

if PYSPARK_AVAILABLE:
    from pyspark.sql import (
        Window,
        functions as sf,
    )
    from pyspark.sql.functions import col
    from pyspark.sql.types import TimestampType


class _BaseFilter(ABC):
    def transform(self, interactions: DataFrameLike) -> DataFrameLike:
        r"""Filter interactions.

        :param interactions: DataFrame with interactions.

        :returns: filtered DataFrame.
        """
        if isinstance(interactions, SparkDataFrame):
            return self._filter_spark(interactions)
        elif isinstance(interactions, PandasDataFrame):
            return self._filter_pandas(interactions)
        elif isinstance(interactions, PolarsDataFrame):
            return self._filter_polars(interactions)
        else:
            msg = f"{self.__class__.__name__} is not implemented for {type(interactions)}"
            raise NotImplementedError(msg)

    @abstractmethod
    def _filter_spark(self, interactions: SparkDataFrame):  # pragma: no cover
        pass

    @abstractmethod
    def _filter_pandas(self, interactions: PandasDataFrame):  # pragma: no cover
        pass

    @abstractmethod
    def _filter_polars(self, interactions: PolarsDataFrame):  # pragma: no cover
        pass


class InteractionEntriesFilter(_BaseFilter):
    """
    Remove interactions less than minimum constraint value and greater
    than maximum constraint value for each column.

    >>> import pandas as pd
    >>> interactions = pd.DataFrame({
    ...    "user_id": [1, 1, 1, 2, 2, 2, 3, 3, 3, 3],
    ...    "item_id": [3, 7, 10, 5, 8, 11, 4, 9, 2, 5],
    ...    "rating": [1, 2, 3, 3, 2, 1, 3, 12, 1, 4]
    ... })
    >>> interactions
        user_id  item_id  rating
    0        1        3       1
    1        1        7       2
    2        1       10       3
    3        2        5       3
    4        2        8       2
    5        2       11       1
    6        3        4       3
    7        3        9      12
    8        3        2       1
    9        3        5       4
    >>> filtered_interactions = InteractionEntriesFilter(min_inter_per_user=4).transform(interactions)
    >>> filtered_interactions
        user_id  item_id  rating
    6        3        4       3
    7        3        9      12
    8        3        2       1
    9        3        5       4
    <BLANKLINE>
    """

    def __init__(
        self,
        query_column: str = "user_id",
        item_column: str = "item_id",
        min_inter_per_user: Optional[int] = None,
        max_inter_per_user: Optional[int] = None,
        min_inter_per_item: Optional[int] = None,
        max_inter_per_item: Optional[int] = None,
        allow_caching: bool = True,
    ):
        r"""
        :param user_column: Name of user interaction column,
            default: ``user_id``.
        :param item_column: Name of item interaction column,
            default: ``item_id``.
        :param min_inter_per_user: Minimum positive value of
            interactions per user. If None, filter doesn't apply,
            default: ``None``.
        :param max_inter_per_user: Maximum positive value of
            interactions per user. If None, filter doesn't apply. Must be
            less than `min_inter_per_user`, default: ``None``.
        :param min_inter_per_item: Minimum positive value of
            interactions per item. If None, filter doesn't apply,
            default: ``None``.
        :param max_inter_per_item: Maximum positive value of
            interactions per item. If None, filter doesn't apply. Must be
            less than `min_inter_per_item`, default: ``None``.
        :param allow_caching: The flag for using caching to optimize calculations.
            default: `True`.
        """
        self.query_column = query_column
        self.item_column = item_column
        self.min_inter_per_user = min_inter_per_user
        self.max_inter_per_user = max_inter_per_user
        self.min_inter_per_item = min_inter_per_item
        self.max_inter_per_item = max_inter_per_item
        self.total_dropped_interactions = 0
        self.allow_caching = allow_caching
        self._sanity_check()

    def _sanity_check(self) -> None:
        if self.min_inter_per_user:
            assert self.min_inter_per_user > 0
        if self.min_inter_per_item:
            assert self.min_inter_per_item > 0
        if self.min_inter_per_user and self.max_inter_per_user:
            assert self.min_inter_per_user < self.max_inter_per_user
        if self.min_inter_per_item and self.max_inter_per_item:
            assert self.min_inter_per_item < self.max_inter_per_item

    def _filter_spark(self, interactions: SparkDataFrame):
        interaction_count = interactions.count()
        return self._iterative_filter(interactions, interaction_count, self._filter_column_spark)

    def _filter_pandas(self, interactions: PandasDataFrame):
        interaction_count = len(interactions)
        return self._iterative_filter(interactions, interaction_count, self._filter_column_pandas)

    def _filter_polars(self, interactions: PolarsDataFrame):
        interaction_count = len(interactions)
        return self._iterative_filter(interactions, interaction_count, self._filter_column_polars)

    def _iterative_filter(self, interactions: DataFrameLike, interaction_count: int, filter_func: Callable):
        is_dropped_user_item = [True, True]
        current_index = 0
        while is_dropped_user_item[0] or is_dropped_user_item[1]:
            if current_index == 0:
                min_inter = self.min_inter_per_user
                max_inter = self.max_inter_per_user
                agg_column = self.query_column
                non_agg_column = self.item_column
            else:
                min_inter = self.min_inter_per_item
                max_inter = self.max_inter_per_item
                agg_column = self.item_column
                non_agg_column = self.query_column

            if min_inter is None and max_inter is None:
                dropped_interact = 0
            else:
                interactions, dropped_interact, interaction_count = filter_func(
                    interactions, interaction_count, agg_column, non_agg_column, min_inter, max_inter
                )
            is_dropped_user_item[current_index] = bool(dropped_interact)
            current_index = (current_index + 1) % 2  # current_index only in (0, 1)

        return interactions

    def _filter_column_pandas(
        self,
        interactions: PandasDataFrame,
        interaction_count: int,
        agg_column: str,
        non_agg_column: str,
        min_inter: Optional[int] = None,
        max_inter: Optional[int] = None,
    ) -> tuple[PandasDataFrame, int, int]:
        filtered_interactions = interactions.copy(deep=True)

        filtered_interactions["count"] = filtered_interactions.groupby(agg_column, sort=False)[
            non_agg_column
        ].transform(len)
        if min_inter:
            filtered_interactions = filtered_interactions[filtered_interactions["count"] >= min_inter]
        if max_inter:
            filtered_interactions = filtered_interactions[filtered_interactions["count"] <= max_inter]
        filtered_interactions.drop(columns=["count"], inplace=True)

        end_len_dataframe = len(filtered_interactions)
        different_len = interaction_count - end_len_dataframe

        return filtered_interactions, different_len, end_len_dataframe

    def _filter_column_spark(
        self,
        interactions: SparkDataFrame,
        interaction_count: int,
        agg_column: str,
        non_agg_column: str,
        min_inter: Optional[int] = None,
        max_inter: Optional[int] = None,
    ) -> tuple[SparkDataFrame, int, int]:
        filtered_interactions = interactions.withColumn(
            "count", sf.count(non_agg_column).over(Window.partitionBy(agg_column))
        )
        if min_inter:
            filtered_interactions = filtered_interactions.filter(sf.col("count") >= min_inter)
        if max_inter:
            filtered_interactions = filtered_interactions.filter(sf.col("count") <= max_inter)
        filtered_interactions = filtered_interactions.drop("count")

        if self.allow_caching:
            filtered_interactions.cache()
            interactions.unpersist()
        end_len_dataframe = filtered_interactions.count()
        different_len = interaction_count - end_len_dataframe

        return filtered_interactions, different_len, end_len_dataframe

    def _filter_column_polars(
        self,
        interactions: PolarsDataFrame,
        interaction_count: int,
        agg_column: str,
        non_agg_column: str,
        min_inter: Optional[int] = None,
        max_inter: Optional[int] = None,
    ) -> tuple[PolarsDataFrame, int, int]:
        filtered_interactions = interactions.with_columns(
            pl.col(non_agg_column).count().over(pl.col(agg_column)).alias("count")
        )
        if min_inter:
            filtered_interactions = filtered_interactions.filter(pl.col("count") >= min_inter)
        if max_inter:
            filtered_interactions = filtered_interactions.filter(pl.col("count") <= max_inter)
        filtered_interactions = filtered_interactions.drop("count")

        end_len_dataframe = len(filtered_interactions)
        different_len = interaction_count - end_len_dataframe

        return filtered_interactions, different_len, end_len_dataframe


class MinCountFilter(_BaseFilter):
    """
    Remove entries with entities (e.g. users, items) which are presented in `interactions`
    less than `num_entries` times. The `interactions` is grouped by `groupby_column`,
    which is entry column name, to calculate counts.

    >>> import pandas as pd
    >>> data_frame = pd.DataFrame({"user_id": [1, 1, 2]})
    >>> MinCountFilter(2).transform(data_frame)
        user_id
    0         1
    1         1
    <BLANKLINE>
    """

    def __init__(
        self,
        num_entries: int,
        groupby_column: str = "user_id",
    ):
        r"""
        :param num_entries: minimal number of times the entry should appear in dataset
            in order to remain.
        :param group_by: entity column, which is used to calculate entity occurrence couns,
            default: ``user_id``.
        """
        self.num_entries = num_entries
        self.groupby_column = groupby_column
        self._sanity_check()

    def _sanity_check(self) -> None:
        assert self.num_entries > 0

    def _filter_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        count_by_group = interactions.groupBy(self.groupby_column).agg(
            sf.count(self.groupby_column).alias(f"{self.groupby_column}_temp_count")
        )
        remaining_entities = count_by_group.filter(
            count_by_group[f"{self.groupby_column}_temp_count"] >= self.num_entries
        ).select(self.groupby_column)

        return interactions.join(remaining_entities, on=self.groupby_column, how="inner")

    def _filter_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        filtered_interactions = interactions.copy(deep=True)
        filtered_interactions["count"] = filtered_interactions.groupby(self.groupby_column)[
            self.groupby_column
        ].transform(len)
        return filtered_interactions[filtered_interactions["count"] >= self.num_entries].drop(columns=["count"])

    def _filter_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        filtered_interactions = interactions.clone()
        count_by_group = (
            filtered_interactions.group_by(self.groupby_column)
            .agg(pl.col(self.groupby_column).count().alias(f"{self.groupby_column}_temp_count"))
            .filter(pl.col(f"{self.groupby_column}_temp_count") >= self.num_entries)
        )
        return filtered_interactions.join(count_by_group, on=self.groupby_column).drop(
            f"{self.groupby_column}_temp_count"
        )


class LowRatingFilter(_BaseFilter):
    """
    Remove records with records less than ``value`` in ``column``.

    >>> import pandas as pd
    >>> data_frame = pd.DataFrame({"rating": [1, 5, 3.5, 4]})
    >>> LowRatingFilter(3.5).transform(data_frame)
         rating
    1       5.0
    2       3.5
    3       4.0
    <BLANKLINE>
    """

    def __init__(
        self,
        value: float,
        rating_column: str = "rating",
    ):
        r"""
        :param value: minimal value the entry should appear in dataset
            in order to remain.
        :param rating_column: the column in which filtering is performed.
        """
        self.value = value
        self.rating_column = rating_column

    def _filter_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        return interactions.filter(interactions[self.rating_column] >= self.value)

    def _filter_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        return interactions[interactions[self.rating_column] >= self.value]

    def _filter_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        return interactions.filter(pl.col(self.rating_column) >= self.value)


class NumInteractionsFilter(_BaseFilter):
    """
    Get first/last ``num_interactions`` interactions for each query.

    >>> import pandas as pd
    >>> from replay.utils.spark_utils import convert2spark
    >>> log_pd = pd.DataFrame({"user_id": ["u1", "u2", "u2", "u3", "u3", "u3"],
    ...                     "item_id": ["i1", "i2","i3", "i1", "i2","i3"],
    ...                     "rating": [1., 0.5, 3, 1, 0, 1],
    ...                     "timestamp": ["2020-01-01 23:59:59", "2020-02-01 00:00:00",
    ...                                   "2020-02-01 00:00:01", "2020-01-01 00:04:15",
    ...                                   "2020-01-02 00:04:14", "2020-01-05 23:59:59"]},
    ...             )
    >>> log_pd["timestamp"] = pd.to_datetime(log_pd["timestamp"], format="ISO8601")
    >>> log_sp = convert2spark(log_pd)
    >>> log_sp.show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u2|     i2|   0.5|2020-02-01 00:00:00|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    |     u3|     i1|   1.0|2020-01-01 00:04:15|
    |     u3|     i2|   0.0|2020-01-02 00:04:14|
    |     u3|     i3|   1.0|2020-01-05 23:59:59|
    +-------+-------+------+-------------------+
    <BLANKLINE>

    Only first interaction:

    >>> NumInteractionsFilter(1, True, item_column='item_id').transform(log_sp).orderBy('user_id').show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u2|     i2|   0.5|2020-02-01 00:00:00|
    |     u3|     i1|   1.0|2020-01-01 00:04:15|
    +-------+-------+------+-------------------+
    <BLANKLINE>

    Only last interaction:

    >>> NumInteractionsFilter(1, False).transform(log_sp).orderBy('user_id').show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    |     u3|     i3|   1.0|2020-01-05 23:59:59|
    +-------+-------+------+-------------------+
    <BLANKLINE>

    >>> NumInteractionsFilter(1, False, item_column='item_id').transform(log_sp).orderBy('user_id').show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    |     u3|     i3|   1.0|2020-01-05 23:59:59|
    +-------+-------+------+-------------------+
    <BLANKLINE>
    """

    def __init__(
        self,
        num_interactions: int = 10,
        first: bool = True,
        query_column: str = "user_id",
        timestamp_column: str = "timestamp",
        item_column: Optional[str] = None,
    ):
        r"""
        :param num_interactions: number of interactions to leave per user.
        :param first: take either first ``num_interactions`` or last.
            default: `True`.
        :param query_column: query column.
            default: ``user_id``.
        :param timestamp_column: timestamp column.
            default: ``timestamp``.
        :param item_column: item column to help sort simultaneous interactions.
            If None, it is ignored.
            default: `None`.
        """
        self.num_interactions = num_interactions
        self.first = first
        self.query_column = query_column
        self.timestamp_column = timestamp_column
        self.item_column = item_column
        self.first = first
        self._sanity_check()

    def _sanity_check(self) -> None:
        assert self.num_interactions >= 0

    def _filter_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        sorting_order = [col(self.timestamp_column)]
        if self.item_column is not None:
            sorting_order.append(col(self.item_column))

        if not self.first:
            sorting_order = [col_.desc() for col_ in sorting_order]

        window = Window().orderBy(*sorting_order).partitionBy(col(self.query_column))

        return (
            interactions.withColumn("temp_rank", sf.row_number().over(window))
            .filter(col("temp_rank") <= self.num_interactions)
            .drop("temp_rank")
        )

    def _filter_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        filtered_interactions = interactions.copy(deep=True)
        sorting_columns = [self.timestamp_column]
        if self.item_column is not None:
            sorting_columns.append(self.item_column)

        ascending = [self.first] * len(sorting_columns)

        filtered_interactions["temp_rank"] = (
            filtered_interactions.sort_values(sorting_columns, ascending=ascending)
            .groupby(self.query_column)
            .cumcount()
        )
        return filtered_interactions[filtered_interactions["temp_rank"] < self.num_interactions].drop(
            columns=["temp_rank"]
        )

    def _filter_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        sorting_columns = [self.timestamp_column]
        if self.item_column is not None:
            sorting_columns.append(self.item_column)

        descending = not self.first

        return (
            interactions.sort(sorting_columns, descending=descending)
            .with_columns(pl.col(self.query_column).cum_count().over(self.query_column).alias("temp_rank"))
            .filter(pl.col("temp_rank") <= self.num_interactions)
            .drop("temp_rank")
        )


class EntityDaysFilter(_BaseFilter):
    """
    Get first/last ``days`` of interactions by entity.

    >>> import pandas as pd
    >>> from replay.utils.spark_utils import convert2spark
    >>> log_pd = pd.DataFrame({"user_id": ["u1", "u2", "u2", "u3", "u3", "u3"],
    ...                     "item_id": ["i1", "i2","i3", "i1", "i2","i3"],
    ...                     "rating": [1., 0.5, 3, 1, 0, 1],
    ...                     "timestamp": ["2020-01-01 23:59:59", "2020-02-01 00:00:00",
    ...                                   "2020-02-01 00:00:01", "2020-01-01 00:04:15",
    ...                                   "2020-01-02 00:04:14", "2020-01-05 23:59:59"]},
    ...             )
    >>> log_pd["timestamp"] = pd.to_datetime(log_pd["timestamp"], format="ISO8601")
    >>> log_sp = convert2spark(log_pd)
    >>> log_sp.orderBy('user_id', 'item_id').show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u2|     i2|   0.5|2020-02-01 00:00:00|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    |     u3|     i1|   1.0|2020-01-01 00:04:15|
    |     u3|     i2|   0.0|2020-01-02 00:04:14|
    |     u3|     i3|   1.0|2020-01-05 23:59:59|
    +-------+-------+------+-------------------+
    <BLANKLINE>

    Get first day by users:

    >>> EntityDaysFilter(1, True, entity_column='user_id').transform(log_sp).orderBy('user_id', 'item_id').show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u2|     i2|   0.5|2020-02-01 00:00:00|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    |     u3|     i1|   1.0|2020-01-01 00:04:15|
    |     u3|     i2|   0.0|2020-01-02 00:04:14|
    +-------+-------+------+-------------------+
    <BLANKLINE>

    Get last day by item:

    >>> EntityDaysFilter(1, False, entity_column='item_id').transform(log_sp).orderBy('item_id', 'user_id').show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u3|     i1|   1.0|2020-01-01 00:04:15|
    |     u2|     i2|   0.5|2020-02-01 00:00:00|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    +-------+-------+------+-------------------+
    <BLANKLINE>
    """

    def __init__(
        self,
        days: int = 10,
        first: bool = True,
        entity_column: str = "user_id",
        timestamp_column: str = "timestamp",
    ):
        r"""
        :param days: how many days to return per entity.
            default: `10`.
        :param first: take either first ``num_interactions`` or last.
            default: `True`.
        :param entity_column: query/item column.
            default: ``user_id``.
        :param timestamp_column: timestamp column.
            default: ``timestamp``.
        """
        self.days = days
        self.first = first
        self.entity_column = entity_column
        self.timestamp_column = timestamp_column
        self.first = first
        self._sanity_check()

    def _sanity_check(self) -> None:
        assert self.days > 0

    def _filter_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        window = Window.partitionBy(self.entity_column)
        if self.first:
            filtered_interactions = (
                interactions.withColumn("min_date", sf.min(col(self.timestamp_column)).over(window))
                .filter(col(self.timestamp_column) < col("min_date") + sf.expr(f"INTERVAL {self.days} days"))
                .drop("min_date")
            )
        else:
            filtered_interactions = (
                interactions.withColumn("max_date", sf.max(col(self.timestamp_column)).over(window))
                .filter(col(self.timestamp_column) > col("max_date") - sf.expr(f"INTERVAL {self.days} days"))
                .drop("max_date")
            )
        return filtered_interactions

    def _filter_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        filtered_interactions = interactions.copy(deep=True)

        if self.first:
            filtered_interactions["min_date"] = filtered_interactions.groupby(self.entity_column)[
                self.timestamp_column
            ].transform(min)
            return filtered_interactions[
                (filtered_interactions[self.timestamp_column] - filtered_interactions["min_date"]).dt.days < self.days
            ].drop(columns=["min_date"])
        filtered_interactions["max_date"] = filtered_interactions.groupby(self.entity_column)[
            self.timestamp_column
        ].transform(max)
        return filtered_interactions[
            (filtered_interactions["max_date"] - filtered_interactions[self.timestamp_column]).dt.days < self.days
        ].drop(columns=["max_date"])

    def _filter_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        if self.first:
            return (
                interactions.with_columns(
                    (
                        pl.col(self.timestamp_column).min().over(pl.col(self.entity_column))
                        + pl.duration(days=self.days)
                    ).alias("min_date")
                )
                .filter(pl.col(self.timestamp_column) < pl.col("min_date"))
                .drop("min_date")
            )
        return (
            interactions.with_columns(
                (
                    pl.col(self.timestamp_column).max().over(pl.col(self.entity_column)) - pl.duration(days=self.days)
                ).alias("max_date")
            )
            .filter(pl.col(self.timestamp_column) > pl.col("max_date"))
            .drop("max_date")
        )


class GlobalDaysFilter(_BaseFilter):
    """
    Select first/last days from ``interactions``.

    >>> import pandas as pd
    >>> from replay.utils.spark_utils import convert2spark
    >>> log_pd = pd.DataFrame({"user_id": ["u1", "u2", "u2", "u3", "u3", "u3"],
    ...                     "item_id": ["i1", "i2","i3", "i1", "i2","i3"],
    ...                     "rating": [1., 0.5, 3, 1, 0, 1],
    ...                     "timestamp": ["2020-01-01 23:59:59", "2020-02-01 00:00:00",
    ...                                   "2020-02-01 00:00:01", "2020-01-01 00:04:15",
    ...                                   "2020-01-02 00:04:14", "2020-01-05 23:59:59"]},
    ...             )
    >>> log_pd["timestamp"] = pd.to_datetime(log_pd["timestamp"], format="ISO8601")
    >>> log_sp = convert2spark(log_pd)
    >>> log_sp.show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u2|     i2|   0.5|2020-02-01 00:00:00|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    |     u3|     i1|   1.0|2020-01-01 00:04:15|
    |     u3|     i2|   0.0|2020-01-02 00:04:14|
    |     u3|     i3|   1.0|2020-01-05 23:59:59|
    +-------+-------+------+-------------------+
    <BLANKLINE>

    >>> GlobalDaysFilter(1).transform(log_sp).show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u3|     i1|   1.0|2020-01-01 00:04:15|
    |     u3|     i2|   0.0|2020-01-02 00:04:14|
    +-------+-------+------+-------------------+
    <BLANKLINE>

    >>> GlobalDaysFilter(1, first=False).transform(log_sp).show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u2|     i2|   0.5|2020-02-01 00:00:00|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    +-------+-------+------+-------------------+
    <BLANKLINE>
    """

    def __init__(
        self,
        days: int = 10,
        first: bool = True,
        timestamp_column: str = "timestamp",
    ):
        r"""
        :param days: length of selected data in days.
            default: `10`.
        :param first: take either first ``day`` or last.
            default: `True`.
        :param timestamp_column: timestamp column.
            default: ``timestamp``.
        """
        self.days = days
        self.first = first
        self.timestamp_column = timestamp_column
        self.first = first
        self._sanity_check()

    def _sanity_check(self) -> None:
        assert self.days > 0

    def _filter_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        if self.first:
            start_date = interactions.agg(sf.min(self.timestamp_column)).first()[0]
            end_date = sf.lit(start_date).cast(TimestampType()) + sf.expr(f"INTERVAL {self.days} days")
            return interactions.filter(col(self.timestamp_column) < end_date)

        end_date = interactions.agg(sf.max(self.timestamp_column)).first()[0]
        start_date = sf.lit(end_date).cast(TimestampType()) - sf.expr(f"INTERVAL {self.days} days")
        return interactions.filter(col(self.timestamp_column) > start_date)

    def _filter_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        filtered_interactions = interactions.copy(deep=True)

        if self.first:
            start_date = filtered_interactions[self.timestamp_column].min()
            return filtered_interactions[
                (filtered_interactions[self.timestamp_column] - start_date).dt.days < self.days
            ]
        end_date = filtered_interactions[self.timestamp_column].max()
        return filtered_interactions[(end_date - filtered_interactions[self.timestamp_column]).dt.days < self.days]

    def _filter_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        if self.first:
            return interactions.filter(
                pl.col(self.timestamp_column) < (pl.col(self.timestamp_column).min() + pl.duration(days=self.days))
            )
        return interactions.filter(
            pl.col(self.timestamp_column) > (pl.col(self.timestamp_column).max() - pl.duration(days=self.days))
        )


class TimePeriodFilter(_BaseFilter):
    """
    Select a part of data between ``[start_date, end_date)``.

    >>> import pandas as pd
    >>> from replay.utils.spark_utils import convert2spark
    >>> log_pd = pd.DataFrame({"user_id": ["u1", "u2", "u2", "u3", "u3", "u3"],
    ...                     "item_id": ["i1", "i2","i3", "i1", "i2","i3"],
    ...                     "rating": [1., 0.5, 3, 1, 0, 1],
    ...                     "timestamp": ["2020-01-01 23:59:59", "2020-02-01 00:00:00",
    ...                                   "2020-02-01 00:00:01", "2020-01-01 00:04:15",
    ...                                   "2020-01-02 00:04:14", "2020-01-05 23:59:59"]},
    ...             )
    >>> log_pd["timestamp"] = pd.to_datetime(log_pd["timestamp"], format="ISO8601")
    >>> log_sp = convert2spark(log_pd)
    >>> log_sp.show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u2|     i2|   0.5|2020-02-01 00:00:00|
    |     u2|     i3|   3.0|2020-02-01 00:00:01|
    |     u3|     i1|   1.0|2020-01-01 00:04:15|
    |     u3|     i2|   0.0|2020-01-02 00:04:14|
    |     u3|     i3|   1.0|2020-01-05 23:59:59|
    +-------+-------+------+-------------------+
    <BLANKLINE>

    >>> TimePeriodFilter(
    ...    start_date="2020-01-01 14:00:00",
    ...    end_date=datetime(2020, 1, 3, 0, 0, 0)
    ... ).transform(log_sp).show()
    +-------+-------+------+-------------------+
    |user_id|item_id|rating|          timestamp|
    +-------+-------+------+-------------------+
    |     u1|     i1|   1.0|2020-01-01 23:59:59|
    |     u3|     i2|   0.0|2020-01-02 00:04:14|
    +-------+-------+------+-------------------+
    <BLANKLINE>
    """

    def __init__(
        self,
        start_date: Optional[Union[str, datetime]] = None,
        end_date: Optional[Union[str, datetime]] = None,
        timestamp_column: str = "timestamp",
        time_column_format: str = "%Y-%m-%d %H:%M:%S",
    ):
        r"""
        :param start_date: datetime or str with format ``time_column_format``.
            default: `None`.
        :param end_date: datetime or str with format ``time_column_format``.
            default: `None`.
        :param timestamp_column: timestamp column.
            default: ``timestamp``.
        """
        self.start_date = self._format_datetime(start_date, time_column_format)
        self.end_date = self._format_datetime(end_date, time_column_format)
        self.timestamp_column = timestamp_column

    def _format_datetime(self, date: Optional[Union[str, datetime]], time_format: str) -> datetime:
        if isinstance(date, str):
            date = datetime.strptime(date, time_format)
        return date

    def _filter_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        if self.start_date is None:
            self.start_date = interactions.agg(sf.min(self.timestamp_column)).first()[0]
        if self.end_date is None:
            self.end_date = interactions.agg(sf.max(self.timestamp_column)).first()[0] + timedelta(seconds=1)

        return interactions.filter(
            (col(self.timestamp_column) >= sf.lit(self.start_date))
            & (col(self.timestamp_column) < sf.lit(self.end_date))
        )

    def _filter_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        if self.start_date is None:
            self.start_date = interactions[self.timestamp_column].min()
        if self.end_date is None:
            self.end_date = interactions[self.timestamp_column].max() + timedelta(seconds=1)

        return interactions[
            (interactions[self.timestamp_column] >= self.start_date)
            & (interactions[self.timestamp_column] < self.end_date)
        ]

    def _filter_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        if self.start_date is None:
            self.start_date = interactions.select(self.timestamp_column).min()[0, 0]
        if self.end_date is None:
            self.end_date = interactions.select(self.timestamp_column).max()[0, 0] + pl.duration(seconds=1)

        return interactions.filter(
            pl.col(self.timestamp_column).is_between(self.start_date, self.end_date, closed="left")
        )


class QuantileItemsFilter(_BaseFilter):
    """
    Filter is aimed on undersampling the interactions dataset.

    Filter algorithm performs undersampling by removing `items_proportion` of interactions
    for each items counts that exceeds the `alpha_quantile` value in distribution. Filter firstly
    removes popular items (items that have most interactions). Filter also keeps the original
    relation of items popularity among each other by removing interactions only in range of
    current item count and quantile count (specified by `alpha_quantile`).

    >>> import pandas as pd
    >>> from replay.utils.spark_utils import convert2spark
    >>> log_pd = pd.DataFrame({
    ...        "user_id": [0, 0, 1, 2, 2, 2, 2],
    ...        "item_id": [0, 2, 1, 1, 2, 2, 2]
    ... })
    >>> log_spark = convert2spark(log_pd)
    >>> log_spark.show()
    +-------+-------+
    |user_id|item_id|
    +-------+-------+
    |      0|      0|
    |      0|      2|
    |      1|      1|
    |      2|      1|
    |      2|      2|
    |      2|      2|
    |      2|      2|
    +-------+-------+
    <BLANKLINE>

    >>> QuantileItemsFilter(query_column="user_id").transform(log_spark).show()
    +-------+-------+
    |user_id|item_id|
    +-------+-------+
    |      0|      0|
    |      1|      1|
    |      2|      1|
    |      2|      2|
    |      2|      2|
    |      0|      2|
    +-------+-------+
    <BLANKLINE>
    """

    def __init__(
        self,
        alpha_quantile: float = 0.99,
        items_proportion: float = 0.5,
        query_column: str = "query_id",
        item_column: str = "item_id",
    ) -> None:
        """
        :param alpha_quantile: Quantile value of items counts distribution to keep unchanged.
            Every items count that exceeds this value will be undersampled.
            Default: ``0.99``.
        :param items_proportion: proportion of items counts to remove for items that
            exceeds `alpha_quantile` value in range of current item count and quantile count
            to make sure we keep original relation between items unchanged.
            Default: ``0.5``.
        :param query_column: query column name.
            Default: ``query_id``.
        :param item_column: item column name.
            Default: ``item_id``.
        """
        if not 0 < alpha_quantile < 1:
            msg = "`alpha_quantile` value must be in (0, 1)"
            raise ValueError(msg)
        if not 0 < items_proportion < 1:
            msg = "`items_proportion` value must be in (0, 1)"
            raise ValueError(msg)

        self.alpha_quantile = alpha_quantile
        self.items_proportion = items_proportion
        self.query_column = query_column
        self.item_column = item_column

    def _filter_pandas(self, df: pd.DataFrame):
        items_distribution = df.groupby(self.item_column).size().reset_index().rename(columns={0: "counts"})
        users_distribution = df.groupby(self.query_column).size().reset_index().rename(columns={0: "counts"})
        count_threshold = items_distribution.loc[:, "counts"].quantile(self.alpha_quantile, interpolation="midpoint")
        df_with_counts = df.merge(items_distribution, how="left", on=self.item_column).merge(
            users_distribution, how="left", on=self.query_column, suffixes=["_items", "_users"]
        )
        long_tail = df_with_counts.loc[df_with_counts["counts_items"] <= count_threshold]
        short_tail = df_with_counts.loc[df_with_counts["counts_items"] > count_threshold]
        short_tail["num_items_to_delete"] = self.items_proportion * (
            short_tail["counts_items"] - long_tail["counts_items"].max()
        )
        short_tail["num_items_to_delete"] = short_tail["num_items_to_delete"].astype("int")
        short_tail = short_tail.sort_values("counts_users", ascending=False)

        def get_mask(x):
            mask = np.ones_like(x)
            threshold = x.iloc[0]
            mask[:threshold] = 0
            return mask

        mask = short_tail.groupby(self.item_column)["num_items_to_delete"].transform(get_mask).astype(bool)
        return pd.concat([long_tail[df.columns], short_tail.loc[mask][df.columns]])

    def _filter_polars(self, df: pl.DataFrame):
        items_distribution = df.group_by(self.item_column).len()
        users_distribution = df.group_by(self.query_column).len()
        count_threshold = items_distribution.select("len").quantile(self.alpha_quantile, "midpoint")["len"][0]
        df_with_counts = (
            df.join(items_distribution, how="left", on=self.item_column).join(
                users_distribution, how="left", on=self.query_column
            )
        ).rename({"len": "counts_items", "len_right": "counts_users"})
        long_tail = df_with_counts.filter(pl.col("counts_items") <= count_threshold)
        short_tail = df_with_counts.filter(pl.col("counts_items") > count_threshold)
        max_long_tail_count = long_tail["counts_items"].max()
        items_to_delete = (
            short_tail.select(
                self.query_column,
                self.item_column,
                self.items_proportion * (pl.col("counts_items") - max_long_tail_count),
            )
            .with_columns(pl.col("literal").cast(pl.Int64).alias("num_items_to_delete"))
            .select(self.item_column, "num_items_to_delete")
            .unique(maintain_order=True)
        )
        short_tail = short_tail.join(items_to_delete, how="left", on=self.item_column).sort(
            "counts_users", descending=True
        )
        short_tail = short_tail.with_columns(index=pl.int_range(short_tail.shape[0]))
        grouped = short_tail.group_by(self.item_column, maintain_order=True).agg(
            pl.col("index"), pl.col("num_items_to_delete")
        )
        grouped = grouped.with_columns(
            pl.col("num_items_to_delete").list.get(0),
            (pl.col("index").list.len() - pl.col("num_items_to_delete").list.get(0)).alias("tail"),
        )
        grouped = grouped.with_columns(pl.col("index").list.tail(pl.col("tail")))
        grouped = grouped.explode("index").select("index")
        short_tail = grouped.join(short_tail, how="left", on="index")
        return pl.concat([long_tail.select(df.columns), short_tail.select(df.columns)])

    def _filter_spark(self, df: SparkDataFrame):
        items_distribution = df.groupBy(self.item_column).agg(sf.count(self.query_column).alias("counts_items"))
        users_distribution = df.groupBy(self.query_column).agg(sf.count(self.item_column).alias("counts_users"))
        count_threshold = items_distribution.toPandas().loc[:, "counts_items"].quantile(self.alpha_quantile, "midpoint")
        df_with_counts = df.join(items_distribution, on=self.item_column).join(users_distribution, on=self.query_column)
        long_tail = df_with_counts.filter(sf.col("counts_items") <= count_threshold)
        short_tail = df_with_counts.filter(sf.col("counts_items") > count_threshold)
        max_long_tail_count = long_tail.agg({"counts_items": "max"}).collect()[0][0]
        items_to_delete = (
            short_tail.withColumn(
                "num_items_to_delete",
                (self.items_proportion * (sf.col("counts_items") - max_long_tail_count)).cast("int"),
            )
            .select(self.item_column, "num_items_to_delete")
            .distinct()
        )
        short_tail = short_tail.join(items_to_delete, on=self.item_column, how="left")
        short_tail = short_tail.withColumn(
            "index", sf.row_number().over(Window.partitionBy(self.item_column).orderBy(sf.col("counts_users").desc()))
        )
        short_tail = short_tail.filter(sf.col("index") > sf.col("num_items_to_delete"))
        return long_tail.select(df.columns).union(short_tail.select(df.columns))


class ConsecutiveDuplicatesFilter(_BaseFilter):
    """Removes consecutive duplicate items from sequential dataset.

    >>> import datetime as dt
    >>> import pandas as pd
    >>> from replay.utils.spark_utils import convert2spark
    >>> interactions = pd.DataFrame({
    ...     "user_id": ["u0", "u1", "u1", "u0", "u0", "u0", "u1", "u0"],
    ...     "item_id": ["i0", "i1", "i1", "i2", "i0", "i1", "i2", "i1"],
    ...     "timestamp": [dt.datetime(2024, 1, 1) + dt.timedelta(days=i) for i in range(8)]
    ... })
    >>> interactions = convert2spark(interactions)
    >>> interactions.show()
    +-------+-------+-------------------+
    |user_id|item_id|          timestamp|
    +-------+-------+-------------------+
    |     u0|     i0|2024-01-01 00:00:00|
    |     u1|     i1|2024-01-02 00:00:00|
    |     u1|     i1|2024-01-03 00:00:00|
    |     u0|     i2|2024-01-04 00:00:00|
    |     u0|     i0|2024-01-05 00:00:00|
    |     u0|     i1|2024-01-06 00:00:00|
    |     u1|     i2|2024-01-07 00:00:00|
    |     u0|     i1|2024-01-08 00:00:00|
    +-------+-------+-------------------+
    <BLANKLINE>

    >>> ConsecutiveDuplicatesFilter(query_column="user_id").transform(interactions).show()
    +-------+-------+-------------------+
    |user_id|item_id|          timestamp|
    +-------+-------+-------------------+
    |     u0|     i0|2024-01-01 00:00:00|
    |     u0|     i2|2024-01-04 00:00:00|
    |     u0|     i0|2024-01-05 00:00:00|
    |     u0|     i1|2024-01-06 00:00:00|
    |     u1|     i1|2024-01-02 00:00:00|
    |     u1|     i2|2024-01-07 00:00:00|
    +-------+-------+-------------------+
    <BLANKLINE>
    """

    def __init__(
        self,
        keep: Literal["first", "last"] = "first",
        query_column: str = "query_id",
        item_column: str = "item_id",
        timestamp_column: str = "timestamp",
    ) -> None:
        """
        :param keep: whether to keep first or last occurrence,
            Default: ``first``.
        :param query_column: query column,
            Default: ``query_id``.
        :param item_column: item column,
            Default: ``item_id``.
        :param timestamp_column: timestamp column,
            Default: ``timestamp``.
        """
        super().__init__()
        self.query_column = query_column
        self.item_column = item_column
        self.timestamp_column = timestamp_column

        if keep not in ("first", "last"):
            msg = "`keep` must be either 'first' or 'last'"
            raise ValueError(msg)

        self.bias = 1 if keep == "first" else -1
        self.temporary_column = f"__shifted_{uuid4().hex[:8]}"

    def _filter_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        interactions = interactions.sort_values(self.timestamp_column)
        interactions[self.temporary_column] = interactions.groupby(self.query_column)[self.item_column].shift(
            periods=self.bias
        )
        return (
            interactions[interactions[self.item_column] != interactions[self.temporary_column]]
            .drop(self.temporary_column, axis=1)
            .reset_index(drop=True)
        )

    def _filter_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        return (
            interactions.sort(self.timestamp_column)
            .with_columns(
                pl.col(self.item_column).shift(n=self.bias).over(self.query_column).alias(self.temporary_column)
            )
            .filter((pl.col(self.item_column) != pl.col(self.temporary_column)).fill_null(True))
            .drop(self.temporary_column)
        )

    def _filter_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        window = Window.partitionBy(self.query_column).orderBy(self.timestamp_column)
        return (
            interactions.withColumn(self.temporary_column, sf.lag(self.item_column, offset=self.bias).over(window))
            .where((sf.col(self.item_column) != sf.col(self.temporary_column)) | sf.col(self.temporary_column).isNull())
            .drop(self.temporary_column)
        )



from typing import Any, Optional

import numpy as np
import pandas as pd
from scipy.sparse import csc_matrix
from sklearn.linear_model import ElasticNet

from replay.data import Dataset
from replay.utils import PYSPARK_AVAILABLE
from replay.utils.session_handler import State
from replay.utils.spark_utils import spark_to_pandas

from .base_neighbour_rec import NeighbourRec
from .extensions.ann.index_builders.base_index_builder import IndexBuilder

if PYSPARK_AVAILABLE:
    from pyspark.sql import types as st


class SLIM(NeighbourRec):
    """`SLIM: Sparse Linear Methods for Top-N Recommender Systems
    <http://glaros.dtc.umn.edu/gkhome/fetch/papers/SLIM2011icdm.pdf>`_"""

    def _get_ann_infer_params(self) -> dict[str, Any]:
        return {
            "features_col": None,
        }

    _search_space = {
        "beta": {"type": "loguniform", "args": [1e-6, 5]},
        "lambda_": {"type": "loguniform", "args": [1e-6, 2]},
    }

    def __init__(
        self,
        beta: float = 0.01,
        lambda_: float = 0.01,
        seed: Optional[int] = None,
        index_builder: Optional[IndexBuilder] = None,
        allow_collect_to_master: bool = False,
    ):
        """
        :param beta: l2 regularization
        :param lambda_: l1 regularization
        :param seed: random seed
        :param index_builder: `IndexBuilder` instance that adds ANN functionality.
            If not set, then ann will not be used.
        :param allow_collect_to_master: Flag allowing spark to make a collection to the master node,
            Default: ``False``.
        """
        self.init_index_builder(index_builder)

        if beta < 0 or lambda_ <= 0:
            msg = "Invalid regularization parameters"
            raise ValueError(msg)
        self.beta = beta
        self.lambda_ = lambda_
        self.seed = seed
        self.allow_collect_to_master = allow_collect_to_master

    @property
    def _init_args(self):
        return {
            "beta": self.beta,
            "lambda_": self.lambda_,
            "seed": self.seed,
            "index_builder": self.index_builder.init_meta_as_dict() if self.index_builder else None,
            "allow_collect_to_master": self.allow_collect_to_master,
        }

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        interactions = dataset.interactions.select(self.query_column, self.item_column, self.rating_column)
        pandas_interactions = spark_to_pandas(interactions, self.allow_collect_to_master)
        interactions_matrix = csc_matrix(
            (
                pandas_interactions[self.rating_column],
                (
                    pandas_interactions[self.query_column],
                    pandas_interactions[self.item_column],
                ),
            ),
            shape=(self._query_dim, self._item_dim),
        )
        similarity = (
            State()
            .session.createDataFrame(pandas_interactions[self.item_column], st.IntegerType())
            .withColumnRenamed("value", "item_idx_one")
        )

        alpha = self.beta + self.lambda_
        l1_ratio = self.lambda_ / alpha

        regression = ElasticNet(
            alpha=alpha,
            l1_ratio=l1_ratio,
            fit_intercept=False,
            max_iter=5000,
            random_state=self.seed,
            selection="random",
            positive=True,
        )

        def slim_column(pandas_df: pd.DataFrame) -> pd.DataFrame:  # pragma: no cover
            """
            fit similarity matrix with ElasticNet
            :param pandas_df: pd.Dataframe
            :return: pd.Dataframe
            """
            idx = int(pandas_df["item_idx_one"][0])
            column = interactions_matrix[:, idx]
            column_arr = column.toarray().ravel()
            interactions_matrix[interactions_matrix[:, idx].nonzero()[0], idx] = 0

            regression.fit(interactions_matrix, column_arr)
            interactions_matrix[:, idx] = column
            good_idx = np.argwhere(regression.coef_ > 0).reshape(-1)
            good_values = regression.coef_[good_idx]
            similarity_row = {
                "item_idx_one": good_idx,
                "item_idx_two": idx,
                "similarity": good_values,
            }
            return pd.DataFrame(data=similarity_row)

        self.similarity = similarity.groupby("item_idx_one").applyInPandas(
            slim_column,
            "item_idx_one int, item_idx_two int, similarity double",
        )
        self.similarity.cache().count()



from dataclasses import dataclass
from typing import Optional

import torch


@dataclass(frozen=True)
class SCEParams:
    """Set of parameters for ScalableCrossEntropyLoss.

    Constructor arguments:
    :param n_buckets: Number of buckets into which samples will be distributed.
    :param bucket_size_x: Number of item hidden representations that will be in each bucket.
    :param bucket_size_y: Number of item embeddings that will be in each bucket.
    :param mix_x: Whether a randomly generated matrix will be multiplied by the model output matrix or not.
        Default: ``False``.
    """

    n_buckets: int
    bucket_size_x: int
    bucket_size_y: int
    mix_x: bool = False

    def _get_not_none_params(self):
        return [self.n_buckets, self.bucket_size_x, self.bucket_size_y]


class ScalableCrossEntropyLoss:
    def __init__(self, sce_params: SCEParams):
        """
        ScalableCrossEntropyLoss for Sequential Recommendations with Large Item Catalogs.
        Reference article may be found at https://arxiv.org/pdf/2409.18721.

        :param SCEParams: Dataclass with ScalableCrossEntropyLoss parameters.
            Dataclass contains following values:
                :param n_buckets: Number of buckets into which samples will be distributed.
                :param bucket_size_x: Number of item hidden representations that will be in each bucket.
                :param bucket_size_y: Number of item embeddings that will be in each bucket.
                :param mix_x: Whether a randomly generated matrix will be multiplied by the model output matrix or not.
                    Default: ``False``.
        """
        assert all(
            param is not None for param in sce_params._get_not_none_params()
        ), "You should define ``n_buckets``, ``bucket_size_x``, ``bucket_size_y`` when using SCE loss function."
        self._n_buckets = sce_params.n_buckets
        self._bucket_size_x = sce_params.bucket_size_x
        self._bucket_size_y = sce_params.bucket_size_y
        self._mix_x = sce_params.mix_x

    def __call__(
        self,
        embeddings: torch.Tensor,
        positive_labels: torch.LongTensor,
        all_embeddings: torch.Tensor,
        padding_mask: torch.BoolTensor,
        tokens_mask: Optional[torch.BoolTensor] = None,
    ) -> torch.Tensor:
        """
        ScalableCrossEntropyLoss computation.

        :param embeddings: Matrix of the last transformer block outputs.
        :param positive_labels: Positive labels.
        :param all_embeddings: Matrix of all item embeddings.
        :param padding_mask: Padding mask.
        :param tokens_mask: Tokens mask (need only for Bert4Rec).
            Default: ``None``.
        """
        masked_tokens = padding_mask if tokens_mask is None else ~(~padding_mask + tokens_mask)

        hd = torch.tensor(embeddings.shape[-1])
        x = embeddings.view(-1, hd)
        y = positive_labels.view(-1)
        w = all_embeddings

        correct_class_logits_ = (x * torch.index_select(w, dim=0, index=y)).sum(dim=1)  # (bs,)

        with torch.no_grad():
            if self._mix_x:
                omega = 1 / torch.sqrt(torch.sqrt(hd)) * torch.randn(x.shape[0], self._n_buckets, device=x.device)
                buckets = omega.T @ x
                del omega
            else:
                buckets = (
                    1 / torch.sqrt(torch.sqrt(hd)) * torch.randn(self._n_buckets, hd, device=x.device)
                )  # (n_b, hd)

        with torch.no_grad():
            x_bucket = buckets @ x.T  # (n_b, hd) x (hd, b) -> (n_b, b)
            x_bucket[:, ~padding_mask.view(-1)] = float("-inf")
            _, top_x_bucket = torch.topk(x_bucket, dim=1, k=self._bucket_size_x)  # (n_b, bs_x)
            del x_bucket

            y_bucket = buckets @ w.T  # (n_b, hd) x (hd, n_cl) -> (n_b, n_cl)

            _, top_y_bucket = torch.topk(y_bucket, dim=1, k=self._bucket_size_y)  # (n_b, bs_y)
            del y_bucket

        x_bucket = torch.gather(x, 0, top_x_bucket.view(-1, 1).expand(-1, hd)).view(
            self._n_buckets, self._bucket_size_x, hd
        )  # (n_b, bs_x, hd)
        y_bucket = torch.gather(w, 0, top_y_bucket.view(-1, 1).expand(-1, hd)).view(
            self._n_buckets, self._bucket_size_y, hd
        )  # (n_b, bs_y, hd)

        wrong_class_logits = x_bucket @ y_bucket.transpose(-1, -2)  # (n_b, bs_x, bs_y)
        mask = (
            torch.index_select(y, dim=0, index=top_x_bucket.view(-1)).view(self._n_buckets, self._bucket_size_x)[
                :, :, None
            ]
            == top_y_bucket[:, None, :]
        )  # (n_b, bs_x, bs_y)
        wrong_class_logits = wrong_class_logits.masked_fill(mask, float("-inf"))  # (n_b, bs_x, bs_y)
        correct_class_logits = torch.index_select(correct_class_logits_, dim=0, index=top_x_bucket.view(-1)).view(
            self._n_buckets, self._bucket_size_x
        )[
            :, :, None
        ]  # (n_b, bs_x, 1)
        logits = torch.cat((wrong_class_logits, correct_class_logits), dim=2)  # (n_b, bs_x, bs_y + 1)

        loss_ = torch.nn.functional.cross_entropy(
            logits.view(-1, logits.shape[-1]),
            (logits.shape[-1] - 1)
            * torch.ones(logits.shape[0] * logits.shape[1], dtype=torch.int64, device=logits.device),
            reduction="none",
        )  # (n_b * bs_x,)
        loss = torch.zeros(x.shape[0], device=x.device, dtype=x.dtype)
        loss.scatter_reduce_(0, top_x_bucket.view(-1), loss_, reduce="amax", include_self=False)
        loss = loss[(loss != 0) & (masked_tokens).view(-1)]
        loss = torch.mean(loss)

        return loss



import logging
from typing import Optional

from replay.models.extensions.ann.index_inferers.base_inferer import IndexInferer
from replay.models.extensions.ann.index_inferers.nmslib_filter_index_inferer import NmslibFilterIndexInferer
from replay.models.extensions.ann.index_inferers.nmslib_index_inferer import NmslibIndexInferer
from replay.utils import SparkDataFrame
from replay.utils.spark_utils import spark_to_pandas

from .base_index_builder import IndexBuilder
from .nmslib_index_builder_mixin import NmslibIndexBuilderMixin

logger = logging.getLogger("replay")


class DriverNmslibIndexBuilder(IndexBuilder):
    """
    Builder that builds nmslib hnsw index on driver.
    """

    def produce_inferer(self, filter_seen_items: bool) -> IndexInferer:
        if filter_seen_items:
            return NmslibFilterIndexInferer(self.index_params, self.index_store)
        else:
            return NmslibIndexInferer(self.index_params, self.index_store)

    def build_index(
        self,
        vectors: SparkDataFrame,
        features_col: str,  # noqa: ARG002
        ids_col: Optional[str] = None,  # noqa: ARG002
    ):
        vectors = spark_to_pandas(vectors, self.allow_collect_to_master)
        NmslibIndexBuilderMixin.build_and_save_index(vectors, self.index_params, self.index_store)



import logging
from collections.abc import Iterator
from typing import Optional

import pandas as pd

from replay.models.extensions.ann.index_inferers.base_inferer import IndexInferer
from replay.models.extensions.ann.index_inferers.nmslib_filter_index_inferer import NmslibFilterIndexInferer
from replay.models.extensions.ann.index_inferers.nmslib_index_inferer import NmslibIndexInferer
from replay.utils import PandasDataFrame, SparkDataFrame

from .base_index_builder import IndexBuilder
from .nmslib_index_builder_mixin import NmslibIndexBuilderMixin

logger = logging.getLogger("replay")


class ExecutorNmslibIndexBuilder(IndexBuilder):
    """
    Builder that build nmslib hnsw index on one executor.
    """

    def produce_inferer(self, filter_seen_items: bool) -> IndexInferer:
        if filter_seen_items:
            return NmslibFilterIndexInferer(self.index_params, self.index_store)
        else:
            return NmslibIndexInferer(self.index_params, self.index_store)

    def make_build_index_udf(self):
        """
        Method returns udf to build nmslib index.
        :return: `build_index_udf` pandas UDF
        """

        index_params = self.index_params
        index_store = self.index_store

        def build_index_udf(iterator: Iterator[PandasDataFrame]):  # pragma: no cover
            """Builds index on executor and writes it to shared disk or hdfs.

            Args:
                iterator: iterates on dataframes with vectors/features

            """
            # We collect all iterator values into one dataframe,
            # because we cannot guarantee that `pdf` will contain rows
            # with the same `item_idx_two`.
            # And therefore we cannot call the `addDataPointBatch` iteratively
            # (in build_and_save_index).
            pdf = pd.concat(list(iterator))

            NmslibIndexBuilderMixin.build_and_save_index(pdf, index_params, index_store)

            yield PandasDataFrame(data={"_success": 1}, index=[0])

        return build_index_udf

    def build_index(
        self,
        vectors: SparkDataFrame,
        features_col: str,  # noqa: ARG002
        ids_col: Optional[str] = None,  # noqa: ARG002
    ):
        # to execution in one executor
        vectors = vectors.repartition(1)

        # this function will build the index
        build_index_udf = self.make_build_index_udf()

        # Here we perform materialization (`.collect()`) to build the hnsw index.
        vectors.select("similarity", "item_idx_one", "item_idx_two").mapInPandas(
            build_index_udf, "_success int"
        ).collect()



from collections.abc import Iterable
from typing import Any, Optional, Union

import numpy as np

from replay.data import Dataset
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

from .base_neighbour_rec import NeighbourRec
from .extensions.ann.index_builders.base_index_builder import IndexBuilder

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql.window import Window


class AssociationRulesItemRec(NeighbourRec):
    """
    Item-to-item recommender based on association rules.
    Calculate pairs confidence, lift and confidence_gain defined as
    confidence(a, b)/confidence(!a, b) to get top-k associated items.
    Predict items for queries using lift, confidence or confidence_gain metrics.

    Forecasts will be based on indicators: lift, confidence or confidence_gain.
    It all depends on your choice.

    During class initialization, you can explicitly specify the metric to be used as
    a similarity for calculating the predict.

    You can change your selection before calling `.predict()` or `.predict_pairs()`
    if you set a new value for the `similarity_metric` parameter.

    Usage of ANN functionality requires only sparse indices and only one `similarity_metric`,
    defined in `__init__` will be available during inference.

    >>> import pandas as pd
    >>> from replay.data.dataset import Dataset, FeatureSchema, FeatureInfo, FeatureHint, FeatureType
    >>> from replay.utils.spark_utils import convert2spark
    >>> data_frame = pd.DataFrame({"user_id": [1, 1, 2, 3], "item_id": [1, 2, 2, 3], "rating": [2, 1, 4, 1]})
    >>> data_frame_for_predict = pd.DataFrame({"user_id": [2], "item_id": [1]})
    >>> data_frame
       user_id  item_id  rating
    0         1         1          2
    1         1         2          1
    2         2         2          4
    3         3         3          1
    >>> interactions = convert2spark(data_frame)
    >>> pred_interactions = convert2spark(data_frame_for_predict)
    >>> feature_schema = FeatureSchema(
    ...     [
    ...         FeatureInfo(
    ...             column="user_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.QUERY_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="item_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.ITEM_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="rating",
    ...             feature_type=FeatureType.NUMERICAL,
    ...             feature_hint=FeatureHint.RATING,
    ...         ),
    ...     ]
    ... )
    >>> train_dataset = Dataset(feature_schema, interactions)
    >>> pred_dataset = Dataset(feature_schema.subset(["user_id", "item_id"]), pred_interactions)
    >>> model = AssociationRulesItemRec(min_item_count=1, min_pair_count=0, session_column="user_id")
    >>> res = model.fit(train_dataset)
    >>> model.similarity.orderBy("item_idx_one").show()
    +------------+------------+----------+----------+----+---------------+
    |item_idx_one|item_idx_two|similarity|confidence|lift|confidence_gain|
    +------------+------------+----------+----------+----+---------------+
    |           1|           2|       1.0|       1.0| 1.5|            2.0|
    |           2|           1|       0.5|       0.5| 1.5|       Infinity|
    +------------+------------+----------+----------+----+---------------+
    >>> model.similarity_metric = "confidence"
    >>> model.predict_pairs(pred_interactions, train_dataset).show()
    +-------+-------+------+
    |user_id|item_id|rating|
    +-------+-------+------+
    |      2|      1|   0.5|
    +-------+-------+------+
    >>> model.similarity_metric = "lift"
    >>> model.predict_pairs(pred_interactions, train_dataset).show()
    +-------+-------+------+
    |user_id|item_id|rating|
    +-------+-------+------+
    |      2|      1|   1.5|
    +-------+-------+------+

    Classical model uses items co-occurrence in sessions for
    confidence, lift and confidence_gain calculation
    but rating could also be passed to the model, e.g.
    if you want to apply time smoothing and treat old sessions as less important.
    In this case all items in sessions should have the same rating.
    """

    def _get_ann_infer_params(self) -> dict[str, Any]:
        return {
            "features_col": None,
        }

    can_predict_item_to_item = True
    item_to_item_metrics: list[str] = ["lift", "confidence", "confidence_gain"]
    similarity: SparkDataFrame
    can_change_metric = True
    _search_space = {
        "min_item_count": {"type": "int", "args": [3, 10]},
        "min_pair_count": {"type": "int", "args": [3, 10]},
        "num_neighbours": {"type": "int", "args": [300, 2000]},
        "use_rating": {"type": "categorical", "args": [True, False]},
        "similarity_metric": {
            "type": "categorical",
            "args": ["confidence", "lift"],
        },
    }

    def __init__(
        self,
        session_column: str,
        min_item_count: int = 5,
        min_pair_count: int = 5,
        num_neighbours: Optional[int] = 1000,
        use_rating: bool = False,
        similarity_metric: str = "confidence",
        index_builder: Optional[IndexBuilder] = None,
    ) -> None:
        """
        :param session_column: name of column to group sessions.
            Items are combined by the ``user_id`` column if ``session_column`` is not defined.
        :param min_item_count: items with fewer sessions will be filtered out
        :param min_pair_count: pairs with fewer sessions will be filtered out
        :param num_neighbours: maximal number of neighbours to save for each item
        :param use_rating: flag to use rating values instead of co-occurrence count
            If true, pair rating in session is minimal rating of item in pair.
            Item rating is sum of rating in all sessions.
        :param similarity_metric: `lift` of 'confidence'
            The metric used as a similarity to calculate the prediction,
            one of [``lift``, ``confidence``, ``confidence_gain``]
        :param index_builder: `IndexBuilder` instance that adds ANN functionality.
            If not set, then ann will not be used.
        """
        self.init_index_builder(index_builder)

        self.session_column = session_column
        self.min_item_count = min_item_count
        self.min_pair_count = min_pair_count
        self.num_neighbours = num_neighbours
        self.use_rating = use_rating
        self.similarity_metric = similarity_metric

    @property
    def _init_args(self):
        return {
            "session_column": self.session_column,
            "min_item_count": self.min_item_count,
            "min_pair_count": self.min_pair_count,
            "num_neighbours": self.num_neighbours,
            "use_rating": self.use_rating,
            "similarity_metric": self.similarity_metric,
            "index_builder": self.index_builder.init_meta_as_dict() if self.index_builder else None,
        }

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        """
        1) Filter interactions items by ``min_item_count`` threshold
        2) Calculate items support, pairs confidence, lift and confidence_gain defined as
            confidence(a, b)/confidence(!a, b).
        """
        interactions = (
            dataset.interactions.withColumn(
                self.rating_column,
                sf.col(self.rating_column) if self.use_rating else sf.lit(1),
            )
            .select(self.session_column, self.item_column, self.rating_column)
            .distinct()
        )
        num_sessions = interactions.select(self.session_column).distinct().count()

        frequent_items_cached = (
            interactions.groupBy(self.item_column)
            .agg(
                sf.count(self.item_column).alias("item_count"),
                sf.sum(self.rating_column).alias("item_rating"),
            )
            .filter(sf.col("item_count") >= self.min_item_count)
            .drop("item_count")
        ).cache()

        frequent_items_interactions = interactions.join(
            frequent_items_cached.select(self.item_column), on=self.item_column
        )

        frequent_item_pairs = (
            frequent_items_interactions.withColumnRenamed(self.item_column, "antecedent")
            .withColumnRenamed(self.rating_column, "antecedent_rel")
            .join(
                frequent_items_interactions.withColumnRenamed(self.session_column, self.session_column + "_cons")
                .withColumnRenamed(self.item_column, "consequent")
                .withColumnRenamed(self.rating_column, "consequent_rel"),
                on=[
                    sf.col(self.session_column) == sf.col(self.session_column + "_cons"),
                    sf.col("antecedent") < sf.col("consequent"),
                ],
            )
            # taking minimal rating of item for pair
            .withColumn(
                self.rating_column,
                sf.least(sf.col("consequent_rel"), sf.col("antecedent_rel")),
            )
            .drop(self.session_column + "_cons", "consequent_rel", "antecedent_rel")
        )

        pairs_count = (
            frequent_item_pairs.groupBy("antecedent", "consequent")
            .agg(
                sf.count("consequent").alias("pair_count"),
                sf.sum(self.rating_column).alias("pair_rating"),
            )
            .filter(sf.col("pair_count") >= self.min_pair_count)
        ).drop("pair_count")

        pairs_metrics = pairs_count.unionByName(
            pairs_count.select(
                sf.col("consequent").alias("antecedent"),
                sf.col("antecedent").alias("consequent"),
                sf.col("pair_rating"),
            )
        )

        pairs_metrics = pairs_metrics.join(
            frequent_items_cached.withColumnRenamed("item_rating", "antecedent_rating"),
            on=[sf.col("antecedent") == sf.col(self.item_column)],
        ).drop(self.item_column)

        pairs_metrics = pairs_metrics.join(
            frequent_items_cached.withColumnRenamed("item_rating", "consequent_rating"),
            on=[sf.col("consequent") == sf.col(self.item_column)],
        ).drop(self.item_column)

        pairs_metrics = pairs_metrics.withColumn(
            "confidence",
            sf.col("pair_rating") / sf.col("antecedent_rating"),
        ).withColumn(
            "lift",
            num_sessions * sf.col("confidence") / sf.col("consequent_rating"),
        )

        if self.num_neighbours is not None:
            pairs_metrics = (
                pairs_metrics.withColumn(
                    "similarity_order",
                    sf.row_number().over(
                        Window.partitionBy("antecedent").orderBy(
                            sf.col("lift").desc(),
                            sf.col("consequent").desc(),
                        )
                    ),
                )
                .filter(sf.col("similarity_order") <= self.num_neighbours)
                .drop("similarity_order")
            )

        self.similarity = pairs_metrics.withColumn(
            "confidence_gain",
            sf.when(
                sf.col("consequent_rating") - sf.col("pair_rating") == 0,
                sf.lit(np.inf),
            ).otherwise(
                sf.col("confidence")
                * (num_sessions - sf.col("antecedent_rating"))
                / (sf.col("consequent_rating") - sf.col("pair_rating"))
            ),
        ).select(
            sf.col("antecedent").alias("item_idx_one"),
            sf.col("consequent").alias("item_idx_two"),
            sf.col(self.similarity_metric).alias("similarity"),
            "confidence",
            "lift",
            "confidence_gain",
        )
        self.similarity.cache().count()
        frequent_items_cached.unpersist()

    @property
    def get_similarity(self):
        """
        Return matrix with calculated confidence, lift and confidence gain.
        :return: association rules measures calculated during ``fit`` stage
        """
        return self.similarity

    def get_nearest_items(
        self,
        items: Union[SparkDataFrame, Iterable],
        k: int,
        metric: str = "lift",
        candidates: Optional[Union[SparkDataFrame, Iterable]] = None,
    ) -> SparkDataFrame:
        """
        Get k most similar items be the `metric` for each of the `items`.

        :param items: spark dataframe or list of item ids to find neighbors
        :param k: number of neighbors
        :param metric: `lift` of 'confidence_gain'
        :param candidates: spark dataframe or list of items
            to consider as similar, e.g. popular/new items. If None,
            all items presented during model training are used.
        :return: dataframe with the most similar items an distance,
            where bigger value means greater similarity.
            spark-dataframe with columns ``[item_id, neighbour_item_id, similarity]``
        """
        if metric not in self.item_to_item_metrics:
            msg = f"Select one of the valid distance metrics: {self.item_to_item_metrics}"
            raise ValueError(msg)

        return self._get_nearest_items_wrap(
            items=items,
            k=k,
            metric=metric,
            candidates=candidates,
        )

    def _get_nearest_items(
        self,
        items: SparkDataFrame,
        metric: Optional[str] = None,  # noqa: ARG002
        candidates: Optional[SparkDataFrame] = None,
    ) -> SparkDataFrame:
        """
        Return metric for all available associated items filtered by `candidates`.

        :param items: items to find associated
        :param metric: `lift` of 'confidence_gain'
        :param candidates: items to consider as candidates
        :return: associated items
        """

        pairs_to_consider = self.similarity
        if candidates is not None:
            pairs_to_consider = self.similarity.join(
                sf.broadcast(candidates.withColumnRenamed(self.item_column, "item_idx_two")),
                on="item_idx_two",
            )

        return pairs_to_consider.join(
            sf.broadcast(items.withColumnRenamed(self.item_column, "item_idx_one")),
            on="item_idx_one",
        )

    @property
    def _dataframes(self):
        return {"similarity": self.similarity}



from replay.utils import PYSPARK_AVAILABLE

if PYSPARK_AVAILABLE:
    from pyspark.sql.types import DoubleType, IntegerType, StructField, StructType, TimestampType


def get_schema(
    query_column: str = "query_id",
    item_column: str = "item_id",
    timestamp_column: str = "timestamp",
    rating_column: str = "rating",
    has_timestamp: bool = True,
    has_rating: bool = True,
):
    """
    Get Spark Schema with query_id, item_id, rating, timestamp columns

    :param query_column: column name with query ids
    :param item_column: column name with item ids
    :param timestamp_column: column name with timestamps
    :param rating_column: column name with ratings
    :param has_rating: flag to add rating to schema
    :param has_timestamp: flag to add tomestamp to schema
    """
    base = [
        StructField(query_column, IntegerType()),
        StructField(item_column, IntegerType()),
    ]
    if has_timestamp:
        base += [StructField(timestamp_column, TimestampType())]
    if has_rating:
        base += [StructField(rating_column, DoubleType())]
    return StructType(base)



from os.path import join
from typing import Optional

from replay.data import Dataset
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

from .base_rec import ItemVectorModel, Recommender

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.ml.recommendation import ALS, ALSModel
    from pyspark.sql.types import DoubleType

    from replay.utils.spark_utils import list_to_vector_udf


class ALSWrap(Recommender, ItemVectorModel):
    """Wrapper for `Spark ALS
    <https://spark.apache.org/docs/latest/api/python/pyspark.mllib.html#pyspark.mllib.recommendation.ALS>`_.
    """

    _seed: Optional[int] = None
    _search_space = {
        "rank": {"type": "loguniform_int", "args": [8, 256]},
    }

    def __init__(
        self,
        rank: int = 10,
        implicit_prefs: bool = True,
        seed: Optional[int] = None,
        num_item_blocks: Optional[int] = None,
        num_query_blocks: Optional[int] = None,
    ):
        """
        :param rank: hidden dimension for the approximate matrix
        :param implicit_prefs: flag to use implicit feedback
        :param seed: random seed
        :param num_item_blocks: number of blocks the items will be partitioned into in order
            to parallelize computation.
            if None then will be init with number of partitions of interactions.
        :param num_query_blocks: number of blocks the queries will be partitioned into in order
            to parallelize computation.
            if None then will be init with number of partitions of interactions.
        """
        self.rank = rank
        self.implicit_prefs = implicit_prefs
        self._seed = seed
        self._num_item_blocks = num_item_blocks
        self._num_query_blocks = num_query_blocks

    @property
    def _init_args(self):
        return {
            "rank": self.rank,
            "implicit_prefs": self.implicit_prefs,
            "seed": self._seed,
        }

    def _save_model(self, path: str, additional_params: Optional[dict] = None):
        super()._save_model(path, additional_params)
        self.model.write().overwrite().save(join(path, "model"))

    def _load_model(self, path: str):
        super()._load_model(path)
        self.model = ALSModel.load(join(path, "model"))
        self.model.itemFactors.cache()
        self.model.userFactors.cache()

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        if self._num_item_blocks is None:
            self._num_item_blocks = dataset.interactions.rdd.getNumPartitions()
        if self._num_query_blocks is None:
            self._num_query_blocks = dataset.interactions.rdd.getNumPartitions()

        self.model = ALS(
            rank=self.rank,
            numItemBlocks=self._num_item_blocks,
            numUserBlocks=self._num_query_blocks,
            userCol=self.query_column,
            itemCol=self.item_column,
            ratingCol=self.rating_column,
            implicitPrefs=self.implicit_prefs,
            seed=self._seed,
            coldStartStrategy="drop",
        ).fit(dataset.interactions)
        self.model.itemFactors.cache()
        self.model.userFactors.cache()
        self.model.itemFactors.count()
        self.model.userFactors.count()

    def _clear_cache(self):
        if hasattr(self, "model"):
            self.model.itemFactors.unpersist()
            self.model.userFactors.unpersist()

    def _predict(
        self,
        dataset: Optional[Dataset],
        k: int,
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> SparkDataFrame:
        if (items.count() == self.fit_items.count()) and (
            items.join(self.fit_items, on=self.item_column, how="inner").count() == self.fit_items.count()
        ):
            max_seen = 0
            if filter_seen_items and dataset is not None:
                max_seen_in_interactions = (
                    dataset.interactions.join(queries, on=self.query_column)
                    .groupBy(self.query_column)
                    .agg(sf.count(self.query_column).alias("num_seen"))
                    .select(sf.max("num_seen"))
                    .first()[0]
                )
                max_seen = max_seen_in_interactions if max_seen_in_interactions is not None else 0

            recs_als = self.model.recommendForUserSubset(queries, k + max_seen)
            return (
                recs_als.withColumn("recommendations", sf.explode("recommendations"))
                .withColumn(self.item_column, sf.col(f"recommendations.{self.item_column}"))
                .withColumn(
                    self.rating_column,
                    sf.col("recommendations.rating").cast(DoubleType()),
                )
                .select(self.query_column, self.item_column, self.rating_column)
            )

        return self._predict_pairs(
            pairs=queries.crossJoin(items).withColumn(self.rating_column, sf.lit(1)),
            dataset=dataset,
        )

    def _predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Optional[Dataset] = None,  # noqa: ARG002
    ) -> SparkDataFrame:
        return (
            self.model.transform(pairs)
            .withColumn(self.rating_column, sf.col("prediction").cast(DoubleType()))
            .drop("prediction")
        )

    def _get_features(
        self, ids: SparkDataFrame, features: Optional[SparkDataFrame]  # noqa: ARG002
    ) -> tuple[Optional[SparkDataFrame], Optional[int]]:
        entity = "user" if self.query_column in ids.columns else "item"
        entity_col = self.query_column if self.query_column in ids.columns else self.item_column

        als_factors = getattr(self.model, f"{entity}Factors")
        als_factors = als_factors.withColumnRenamed("id", entity_col).withColumnRenamed("features", f"{entity}_factors")
        return (
            als_factors.join(ids, how="right", on=entity_col),
            self.model.rank,
        )

    def _get_item_vectors(self):
        return self.model.itemFactors.select(
            sf.col("id").alias(self.item_column),
            list_to_vector_udf(sf.col("features")).alias("item_vector"),
        )



"""
This splitter split data by two columns.
"""

from typing import Optional

import numpy as np
import polars as pl

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_splitter import Splitter, SplitterReturnType

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql import Window


class TwoStageSplitter(Splitter):
    """
    Split data by two columns.
    First step: takes `first_divide_size` distinct values of `first_divide_column` to the test split.
    Second step: takes `second_divide_size` of `second_divide_column` among the data
    provided after first step to the test split.

    Example:

    >>> from replay.utils.session_handler import get_spark_session, State
    >>> spark = get_spark_session(1, 1)
    >>> state = State(spark)

    >>> from replay.splitters import TwoStageSplitter
    >>> import pandas as pd
    >>> data_frame = pd.DataFrame({"query_id": [1,1,1,2,2,2],
    ...    "item_id": [1,2,3,1,2,3],
    ...    "relevance": [1,2,3,4,5,6],
    ...    "timestamp": [1,2,3,3,2,1]})
    >>> data_frame
       query_id  item_id  relevance  timestamp
    0         1         1          1          1
    1         1         2          2          2
    2         1         3          3          3
    3         2         1          4          3
    4         2         2          5          2
    5         2         3          6          1
    >>> train, test = TwoStageSplitter(first_divide_size=1, second_divide_size=2, seed=42).split(data_frame)
    >>> test
       query_id  item_id  relevance  timestamp
    3         2         1          4          3
    4         2         2          5          2

    >>> train, test = TwoStageSplitter(first_divide_size=0.5, second_divide_size=2, seed=42).split(data_frame)
    >>> test
       query_id  item_id  relevance  timestamp
    3         2         1          4          3
    4         2         2          5          2

    >>> train, test = TwoStageSplitter(first_divide_size=0.5, second_divide_size=0.7, seed=42).split(data_frame)
    >>> test
       query_id  item_id  relevance  timestamp
    3         2         1          4          3
    4         2         2          5          2
    """

    _init_arg_names = [
        "first_divide_size",
        "second_divide_size",
        "first_divide_column",
        "second_divide_column",
        "shuffle",
        "drop_cold_users",
        "drop_cold_items",
        "seed",
        "query_column",
        "item_column",
        "timestamp_column",
    ]

    def __init__(
        self,
        first_divide_size: float,
        second_divide_size: float,
        first_divide_column: str = "query_id",
        second_divide_column: str = "item_id",
        shuffle=False,
        drop_cold_items: bool = False,
        drop_cold_users: bool = False,
        seed: Optional[int] = None,
        query_column: str = "query_id",
        item_column: Optional[str] = "item_id",
        timestamp_column: Optional[str] = "timestamp",
    ):
        """
        :param second_divide_size: fraction or a number of items per user
        :param first_divide_size: similar to ``item_test_size``,
            but corresponds to the number of users.
            ``None`` is all available users.
        :param shuffle: take random items and not last based on ``timestamp``.
        :param drop_cold_items: flag to drop cold items from test
        :param drop_cold_users: flag to drop cold users from test
        :param seed: random seed
        :param query_column: query id column name
        :param item_column: item id column name
        :param timestamp_column: timestamp column name
        """
        super().__init__(
            drop_cold_items=drop_cold_items,
            drop_cold_users=drop_cold_users,
            query_column=query_column,
            item_column=item_column,
            timestamp_column=timestamp_column,
        )
        self.first_divide_column = first_divide_column
        self.second_divide_column = second_divide_column
        self.first_divide_size = first_divide_size
        self.second_divide_size = second_divide_size
        self.shuffle = shuffle
        self.seed = seed

    def _get_test_values(
        self,
        interactions: DataFrameLike,
    ) -> DataFrameLike:
        """
        :param interactions: input DataFrame
        :return: DataFrame with single column `first_divide_column`
        """
        if isinstance(interactions, SparkDataFrame):
            all_values = interactions.select(self.first_divide_column).distinct().sort(self.first_divide_column)
            user_count = all_values.count()
        elif isinstance(interactions, PandasDataFrame):
            all_values = PandasDataFrame(
                np.sort(interactions[self.first_divide_column].unique()), columns=[self.first_divide_column]
            )
            user_count = len(all_values)
        else:
            all_values = interactions.select(self.first_divide_column).unique().sort(self.first_divide_column)
            user_count = len(all_values)

        value_error = False
        if isinstance(self.first_divide_size, int):
            if 1 <= self.first_divide_size < user_count:
                test_user_count = self.first_divide_size
            else:
                value_error = True
        else:
            if 1 > self.first_divide_size > 0:
                test_user_count = user_count * self.first_divide_size
            else:
                value_error = True
        if value_error:
            msg = f"Invalid value for user_test_size: {self.first_divide_size}"
            raise ValueError(msg)
        if isinstance(interactions, SparkDataFrame):
            test_users = (
                all_values.withColumn("_rand", sf.rand(self.seed))
                .withColumn("_row_num", sf.row_number().over(Window.partitionBy(sf.lit(0)).orderBy("_rand")))
                .filter(f"_row_num <= {test_user_count}")
                .drop("_rand", "_row_num")
            )
        elif isinstance(interactions, PandasDataFrame):
            test_users = all_values.sample(n=int(test_user_count), random_state=self.seed)
        else:
            test_users = all_values.sample(n=int(test_user_count), shuffle=True, seed=self.seed)

        return test_users

    def _split_proportion_spark(self, interactions: SparkDataFrame) -> tuple[SparkDataFrame, SparkDataFrame]:
        counts = interactions.groupBy(self.first_divide_column).count()
        test_users = self._get_test_values(interactions).withColumn("is_test", sf.lit(True))
        if self.shuffle:
            res = self._add_random_partition_spark(
                interactions.join(test_users, how="left", on=self.first_divide_column)
            )
        else:
            res = self._add_time_partition_spark(
                interactions.join(test_users, how="left", on=self.first_divide_column),
                query_column=self.query_column,
            )

        res = res.join(counts, on=self.first_divide_column, how="left")
        res = res.withColumn("_frac", sf.col("_row_num") / sf.col("count"))
        res = res.na.fill({"is_test": False})

        train = res.filter(
            f"""
                    _frac > {self.second_divide_size} OR
                    NOT is_test
                """
        ).drop("_rand", "_row_num", "count", "_frac", "is_test")
        test = res.filter(
            f"""
                    _frac <= {self.second_divide_size} AND
                    is_test
                """
        ).drop("_rand", "_row_num", "count", "_frac", "is_test")

        return train, test

    def _split_proportion_pandas(self, interactions: PandasDataFrame) -> tuple[PandasDataFrame, PandasDataFrame]:
        counts = (
            interactions.groupby(self.first_divide_column).agg(count=(self.first_divide_column, "count")).reset_index()
        )
        test_users = self._get_test_values(interactions)
        test_users["is_test"] = True
        if self.shuffle:
            res = self._add_random_partition_pandas(
                interactions.merge(test_users, how="left", on=self.first_divide_column)
            )
        else:
            res = self._add_time_partition_pandas(
                interactions.merge(test_users, how="left", on=self.first_divide_column),
                query_column=self.query_column,
            )
        res["is_test"].fillna(False, inplace=True)
        res = res.merge(counts, on=self.first_divide_column, how="left")
        res["_frac"] = res["_row_num"] / res["count"]
        train = res[(res["_frac"] > self.second_divide_size) | (~res["is_test"])].drop(
            columns=["_row_num", "count", "_frac", "is_test"]
        )
        test = res[(res["_frac"] <= self.second_divide_size) & (res["is_test"])].drop(
            columns=["_row_num", "count", "_frac", "is_test"]
        )

        return train, test

    def _split_proportion_polars(self, interactions: PolarsDataFrame) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        counts = interactions.group_by(self.first_divide_column).count()
        test_users = self._get_test_values(interactions).with_columns(pl.lit(True).alias("is_test"))
        if self.shuffle:
            res = self._add_random_partition_polars(
                interactions.join(test_users, how="left", on=self.first_divide_column)
            )
        else:
            res = self._add_time_partition_polars(
                interactions.join(test_users, how="left", on=self.first_divide_column),
                query_column=self.query_column,
            )

        res = res.join(counts, on=self.first_divide_column, how="left")
        res = res.with_columns((pl.col("_row_num") / pl.col("count")).alias("_frac"))
        res = res.fill_null(False)

        train = res.filter((pl.col("_frac") > self.second_divide_size) | (~pl.col("is_test"))).drop(
            "_row_num", "count", "_frac", "is_test"
        )
        test = res.filter((pl.col("_frac") <= self.second_divide_size) & pl.col("is_test")).drop(
            "_row_num", "count", "_frac", "is_test"
        )

        return train, test

    def _split_proportion(self, interactions: DataFrameLike) -> SplitterReturnType:
        """
        Proportionate split

        :param interactions: input DataFrame `[self.first_divide_column, self.item_column, self.date_column, relevance]`
        :return: train and test DataFrames
        """
        if isinstance(interactions, SparkDataFrame):
            return self._split_proportion_spark(interactions)
        if isinstance(interactions, PandasDataFrame):
            return self._split_proportion_pandas(interactions)
        if isinstance(interactions, PolarsDataFrame):
            return self._split_proportion_polars(interactions)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)

    def _split_quantity_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        test_users = self._get_test_values(interactions).withColumn("is_test", sf.lit(True))
        if self.shuffle:
            res = self._add_random_partition_spark(
                interactions.join(test_users, how="left", on=self.first_divide_column)
            )
        else:
            res = self._add_time_partition_spark(
                interactions.join(test_users, how="left", on=self.first_divide_column),
                query_column=self.query_column,
            )
        res = res.na.fill({"is_test": False})
        train = res.filter(
            f"""
                    _row_num > {self.second_divide_size} OR
                    NOT is_test
                """
        ).drop("_rand", "_row_num", "is_test")
        test = res.filter(
            f"""
                    _row_num <= {self.second_divide_size} AND
                    is_test
                """
        ).drop("_rand", "_row_num", "is_test")

        return train, test

    def _split_quantity_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        test_users = self._get_test_values(interactions)
        test_users["is_test"] = True
        if self.shuffle:
            res = self._add_random_partition_pandas(
                interactions.merge(test_users, how="left", on=self.first_divide_column)
            )
        else:
            res = self._add_time_partition_pandas(
                interactions.merge(test_users, how="left", on=self.first_divide_column),
                query_column=self.query_column,
            )
        res["is_test"].fillna(False, inplace=True)
        train = res[(res["_row_num"] > self.second_divide_size) | (~res["is_test"])].drop(
            columns=["_row_num", "is_test"]
        )
        test = res[(res["_row_num"] <= self.second_divide_size) & (res["is_test"])].drop(
            columns=["_row_num", "is_test"]
        )

        return train, test

    def _split_quantity_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        test_users = self._get_test_values(interactions).with_columns(pl.lit(True).alias("is_test"))
        if self.shuffle:
            res = self._add_random_partition_polars(
                interactions.join(test_users, how="left", on=self.first_divide_column)
            )
        else:
            res = self._add_time_partition_polars(
                interactions.join(test_users, how="left", on=self.first_divide_column),
                query_column=self.query_column,
            )

        res = res.fill_null(False)
        train = res.filter((pl.col("_row_num") > self.second_divide_size) | (~pl.col("is_test"))).drop(
            "_row_num", "is_test"
        )
        test = res.filter((pl.col("_row_num") <= self.second_divide_size) & pl.col("is_test")).drop(
            "_row_num", "is_test"
        )

        return train, test

    def _split_quantity(self, interactions: DataFrameLike) -> SplitterReturnType:
        """
        Split by quantity

        :param interactions: input DataFrame `[self.first_divide_column, self.item_column, self.date_column, relevance]`
        :return: train and test DataFrames
        """
        if isinstance(interactions, SparkDataFrame):
            return self._split_quantity_spark(interactions)
        if isinstance(interactions, PandasDataFrame):
            return self._split_quantity_pandas(interactions)
        if isinstance(interactions, PolarsDataFrame):
            return self._split_quantity_polars(interactions)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)

    def _core_split(self, interactions: DataFrameLike) -> SplitterReturnType:
        if 0 <= self.second_divide_size < 1.0:
            train, test = self._split_proportion(interactions)
        elif self.second_divide_size >= 1 and isinstance(self.second_divide_size, int):
            train, test = self._split_quantity(interactions)
        else:
            msg = f"`test_size` value must be [0, 1) or a positive integer; test_size={self.second_divide_size}"
            raise ValueError(msg)

        return train, test

    def _add_random_partition_spark(self, dataframe: SparkDataFrame) -> SparkDataFrame:
        """
        Adds `_rand` column and a user index column `_row_num` based on `_rand`.

        :param dataframe: input DataFrame with `query_id` column
        :returns: processed DataFrame
        """
        dataframe = dataframe.withColumn("_rand", sf.rand(self.seed))
        dataframe = dataframe.withColumn(
            "_row_num",
            sf.row_number().over(Window.partitionBy(self.first_divide_column).orderBy("_rand")),
        )
        return dataframe

    def _add_random_partition_pandas(self, dataframe: PandasDataFrame) -> PandasDataFrame:
        res = dataframe.sample(frac=1, random_state=self.seed).sort_values(self.first_divide_column)
        res["_row_num"] = res.groupby(self.first_divide_column, sort=False).cumcount() + 1

        return res

    def _add_random_partition_polars(self, dataframe: PolarsDataFrame) -> PolarsDataFrame:
        res = dataframe.sample(fraction=1, shuffle=True, seed=self.seed).with_columns(
            pl.cum_count(self.first_divide_column).over(self.first_divide_column).alias("_row_num")
        )
        return res

    @staticmethod
    def _add_time_partition_spark(
        dataframe: SparkDataFrame,
        query_column: str = "query_id",
        date_column: str = "timestamp",
    ) -> SparkDataFrame:
        """
        Adds user index `_row_num` based on `timestamp`.

        :param dataframe: input DataFrame with `[timestamp, query_id]`
        :param query_column: user id column name
        :param date_column: timestamp column name
        :returns: processed DataFrame
        """
        res = dataframe.withColumn(
            "_row_num",
            sf.row_number().over(Window.partitionBy(query_column).orderBy(sf.col(date_column).desc())),
        )
        return res

    @staticmethod
    def _add_time_partition_pandas(
        dataframe: PandasDataFrame,
        query_column: str = "query_id",
        date_column: str = "timestamp",
    ) -> PandasDataFrame:
        res = dataframe.copy(deep=True)
        res.sort_values([query_column, date_column], ascending=[True, False], inplace=True)
        res["_row_num"] = res.groupby(query_column, sort=False).cumcount() + 1
        return res

    @staticmethod
    def _add_time_partition_polars(
        dataframe: PolarsDataFrame,
        query_column: str = "query_id",
        date_column: str = "timestamp",
    ) -> PolarsDataFrame:
        res = dataframe.sort(date_column, descending=True).with_columns(
            pl.cum_count(query_column).over(query_column).alias("_row_num")
        )
        return res



from .base_metric import Metric


class HitRate(Metric):
    """
    Percentage of users that have at least one correctly recommended item\
        among top-k.

    .. math::
        HitRate@K(i) = \\max_{j \\in [1..K]}\\mathbb{1}_{r_{ij}}

    .. math::
        HitRate@K = \\frac {\\sum_{i=1}^{N}HitRate@K(i)}{N}

    :math:`\\mathbb{1}_{r_{ij}}` -- indicator function stating that user :math:`i` interacted with item :math:`j`

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> groundtruth
       query_id  item_id
    0         1        5
    1         1        6
    2         1        7
    3         1        8
    4         1        9
    5         1       10
    6         2        6
    7         2        7
    8         2        4
    9         2       10
    10        2       11
    11        3        1
    12        3        2
    13        3        3
    14        3        4
    15        3        5
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> HitRate(2)(recommendations, groundtruth)
    {'HitRate@2': 0.6666666666666666}
    >>> HitRate(2, mode=PerUser())(recommendations, groundtruth)
    {'HitRate-PerUser@2': {1: 1.0, 2: 0.0, 3: 1.0}}
    >>> HitRate(2, mode=Median())(recommendations, groundtruth)
    {'HitRate-Median@2': 1.0}
    >>> HitRate(2, mode=ConfidenceInterval(alpha=0.95))(recommendations, groundtruth)
    {'HitRate-ConfidenceInterval@2': 0.6533213281800181}
    <BLANKLINE>
    """

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], ground_truth: list, pred: list) -> list[float]:
        if not ground_truth or not pred:
            return [0.0 for _ in ks]
        set_gt = set(ground_truth)
        res = []
        for k in ks:
            if len(set(pred[:k]).intersection(set_gt)) > 0:
                res.append(1.0)
            else:
                res.append(0.0)
        return res



from typing import Optional

from replay.utils import DataFrameLike, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_splitter import Splitter, SplitterReturnType


class RandomSplitter(Splitter):
    """Assign records into train and test at random."""

    _init_arg_names = [
        "test_size",
        "drop_cold_users",
        "drop_cold_items",
        "seed",
        "query_column",
        "item_column",
    ]

    def __init__(
        self,
        test_size: float,
        drop_cold_items: bool = False,
        drop_cold_users: bool = False,
        seed: Optional[int] = None,
        query_column: str = "query_id",
        item_column: str = "item_id",
    ):
        """
        :param test_size: test size 0 to 1
        :param drop_cold_items: flag to drop cold items from test
        :param drop_cold_users: flag to drop cold users from test
        :param seed: random seed
        :param query_column: Name of query interaction column
        :param item_column: Name of item interaction column
        """
        super().__init__(
            drop_cold_items=drop_cold_items,
            drop_cold_users=drop_cold_users,
            query_column=query_column,
            item_column=item_column,
        )
        self.seed = seed
        if test_size < 0 or test_size > 1:
            msg = "test_size must between 0 and 1"
            raise ValueError(msg)
        self.test_size = test_size

    def _random_split_spark(
        self, interactions: SparkDataFrame, threshold: float
    ) -> tuple[SparkDataFrame, SparkDataFrame]:
        train, test = interactions.randomSplit([1 - threshold, threshold], self.seed)
        return train, test

    def _random_split_pandas(
        self, interactions: PandasDataFrame, threshold: float
    ) -> tuple[PandasDataFrame, PandasDataFrame]:
        train = interactions.sample(frac=(1 - threshold), random_state=self.seed)
        test = interactions.drop(train.index)
        return train, test

    def _random_split_polars(
        self, interactions: PolarsDataFrame, threshold: float
    ) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        train_size = int(len(interactions) * (1 - threshold)) + 1
        shuffled_interactions = interactions.sample(fraction=1, shuffle=True, seed=self.seed)
        train = shuffled_interactions[:train_size]
        test = shuffled_interactions[train_size:]
        return train, test

    def _core_split(self, interactions: DataFrameLike) -> SplitterReturnType:
        if isinstance(interactions, SparkDataFrame):
            return self._random_split_spark(interactions, self.test_size)
        if isinstance(interactions, PandasDataFrame):
            return self._random_split_pandas(interactions, self.test_size)
        if isinstance(interactions, PolarsDataFrame):
            return self._random_split_polars(interactions, self.test_size)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)



import math

from .base_metric import Metric


class NDCG(Metric):
    """
    Normalized Discounted Cumulative Gain is a metric
    that takes into account positions of relevant items.

    This is the binary version, it takes into account
    whether the item was consumed or not, relevance value is ignored.

    .. math::
        DCG@K(i) = \\sum_{j=1}^{K}\\frac{\\mathbb{1}_{r_{ij}}}{\\log_2 (j+1)}


    :math:`\\mathbb{1}_{r_{ij}}` -- indicator function showing that user :math:`i` interacted with item :math:`j`

    To get from :math:`DCG` to :math:`nDCG` we calculate the biggest possible value of `DCG`
    for user :math:`i` and recommendation length :math:`K`.

    .. math::
        IDCG@K(i) = max(DCG@K(i)) = \\sum_{j=1}^{K}\\frac{\\mathbb{1}_{j\\le|Rel_i|}}{\\log_2 (j+1)}

    .. math::
        nDCG@K(i) = \\frac {DCG@K(i)}{IDCG@K(i)}

    :math:`|Rel_i|` -- number of relevant items for user :math:`i`

    Metric is averaged by users.

    .. math::
        nDCG@K = \\frac {\\sum_{i=1}^{N}nDCG@K(i)}{N}

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> groundtruth
       query_id  item_id
    0         1        5
    1         1        6
    2         1        7
    3         1        8
    4         1        9
    5         1       10
    6         2        6
    7         2        7
    8         2        4
    9         2       10
    10        2       11
    11        3        1
    12        3        2
    13        3        3
    14        3        4
    15        3        5
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> NDCG(2)(recommendations, groundtruth)
    {'NDCG@2': 0.3333333333333333}
    >>> NDCG(2, mode=PerUser())(recommendations, groundtruth)
    {'NDCG-PerUser@2': {1: 0.38685280723454163, 2: 0.0, 3: 0.6131471927654584}}
    >>> NDCG(2, mode=Median())(recommendations, groundtruth)
    {'NDCG-Median@2': 0.38685280723454163}
    >>> NDCG(2, mode=ConfidenceInterval(alpha=0.95))(recommendations, groundtruth)
    {'NDCG-ConfidenceInterval@2': 0.3508565839953337}
    <BLANKLINE>
    """

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], ground_truth: list, pred: list) -> list[float]:
        if not pred or not ground_truth:
            return [0.0 for _ in ks]
        set_gt = set(ground_truth)
        res = []
        for k in ks:
            pred_len = min(k, len(pred))
            ground_truth_len = min(k, len(ground_truth))
            denom = [1 / math.log2(i + 2) for i in range(k)]
            dcg = sum(denom[i] for i in range(pred_len) if pred[i] in set_gt)
            idcg = sum(denom[:ground_truth_len])
            res.append(dcg / idcg)
        return res



"""
This class calculates loss function for optimization process
"""

import collections
import logging
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Union

from replay.metrics import Metric
from replay.utils import PYSPARK_AVAILABLE, SparkDataFrame

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf

if TYPE_CHECKING:
    from optuna import Trial


SplitData = collections.namedtuple(  # noqa: PYI024
    "SplitData",
    "train_dataset test_dataset queries items",
)


class ObjectiveWrapper:
    """
    This class is implemented according to
    `instruction <https://optuna.readthedocs.io/en/stable/faq.html#how-to-define-objective-functions-that-have-own-arguments>`_
    on integration with ``optuna``.

    Criterion is calculated with ``__call__``,
    other arguments are passed into ``__init__``.
    """

    def __init__(self, objective_calculator: Callable[..., float], **kwargs: Any):
        self.objective_calculator = objective_calculator
        self.kwargs = kwargs

    def __call__(self, trial: "Trial") -> float:
        """
        Calculate criterion for ``optuna``.

        :param trial: current trial
        :return: criterion value
        """
        return self.objective_calculator(trial=trial, **self.kwargs)


def suggest_params(
    trial: "Trial",
    search_space: dict[str, dict[str, Union[str, list]]],
) -> dict:
    """
    This function suggests params to try.

    :param trial: optuna trial
    :param search_space: hyper parameters and their bounds
    :return: dict with parameter values
    """
    suggest_dict = {
        "uniform": trial.suggest_uniform,
        "int": trial.suggest_int,
        "loguniform": trial.suggest_loguniform,
        "loguniform_int": partial(trial.suggest_int, log=True),
    }

    res = {}
    for param in search_space:
        border = search_space[param]["args"]
        param_type = search_space[param]["type"]
        if param_type == "categorical":
            res[param] = trial.suggest_categorical(param, border)
        else:
            low, high = border
            suggest_fn = suggest_dict[param_type]
            res[param] = suggest_fn(param, low=low, high=high)
    return res


def calculate_criterion_value(
    criterion: Metric, recommendations: SparkDataFrame, ground_truth: SparkDataFrame
) -> float:
    """
    Calculate criterion value for given parameters
    :param criterion: optimization metric
    :param recommendations: calculated recommendations
    :param ground_truth: test data
    :return: criterion value
    """
    result_dict = criterion(recommendations, ground_truth)
    return next(iter(result_dict.values()))


def eval_quality(
    split_data: SplitData,
    recommender,
    criterion: Metric,
    k: int,
) -> float:
    """
    Calculate criterion value using model, data and criterion parameters
    :param split_data: data to train and test model
    :param recommender: recommender model
    :param criterion: optimization metric
    :param k: length of a recommendation list
    :return: criterion value
    """
    logger = logging.getLogger("replay")
    logger.debug("Fitting model inside optimization")
    recommender._fit_wrap(
        split_data.train_dataset,
    )
    logger.debug("Predicting inside optimization")
    recs = recommender._predict_wrap(
        dataset=split_data.train_dataset,
        k=k,
        queries=split_data.queries,
        items=split_data.items,
    )
    logger.debug("Calculating criterion")
    criterion_value = calculate_criterion_value(criterion, recs, split_data.test_dataset.interactions)
    logger.debug("%s=%.6f", criterion, criterion_value)
    return criterion_value


def scenario_objective_calculator(
    trial: "Trial",
    search_space: dict[str, list],
    split_data: SplitData,
    recommender,
    criterion: Metric,
    k: int,
) -> float:
    """
    Sample parameters and calculate criterion value
    :param trial: optuna trial
    :param search_space: hyper parameter search space
    :param split_data: data to train and test model
    :param recommender: recommender model
    :param criterion: optimization metric
    :param k: length of a recommendation list
    :return: criterion value
    """
    params_for_trial = suggest_params(trial, search_space)
    recommender.set_params(**params_for_trial)
    return eval_quality(split_data, recommender, criterion, k)


class ItemKNNObjective:
    """
    This class is implemented according to
    `instruction <https://optuna.readthedocs.io/en/stable/faq.html#how-to-define-objective-functions-that-have-own-arguments>`_
    on integration with ``optuna``.

    Criterion is calculated with ``__call__``,
    other arguments are passed into ``__init__``.
    """

    def __init__(self, **kwargs: Any):
        self.kwargs = kwargs
        max_neighbours = self.kwargs["search_space"]["num_neighbours"]["args"][1]
        model = self.kwargs["recommender"]
        split_data = self.kwargs["split_data"]
        train_dataset = split_data.train_dataset
        model.num_neighbours = max_neighbours

        self.query_column = train_dataset.feature_schema.query_id_column
        self.item_column = train_dataset.feature_schema.item_id_column
        self.rating_column = train_dataset.feature_schema.interactions_rating_column
        self.timestamp_col = train_dataset.feature_schema.interactions_timestamp_column

        df = train_dataset.interactions.select(self.query_column, self.item_column, self.rating_column)
        if not model.use_rating:
            df = df.withColumn(self.rating_column, sf.lit(1))

        self.dot_products = model._get_products(df).cache()

    def objective_calculator(
        self,
        trial: "Trial",
        search_space: dict[str, list],
        split_data: SplitData,
        recommender,
        criterion: Metric,
        k: int,
    ) -> float:
        """
        Sample parameters and calculate criterion value
        :param trial: optuna trial
        :param search_space: hyper parameter search space
        :param split_data: data to train and test model
        :param recommender: recommender model
        :param criterion: optimization metric
        :param k: length of a recommendation list
        :return: criterion value
        """
        params_for_trial = suggest_params(trial, search_space)
        recommender.set_params(**params_for_trial)
        recommender.fit_queries = split_data.train_dataset.interactions.select(self.query_column).distinct()
        recommender.fit_items = split_data.train_dataset.interactions.select(self.item_column).distinct()
        similarity = recommender._shrink(self.dot_products, recommender.shrink)
        recommender.similarity = recommender._get_k_most_similar(similarity).cache()
        recs = recommender._predict_wrap(
            dataset=split_data.train_dataset,
            k=k,
            queries=split_data.queries,
            items=split_data.items,
        )
        logger = logging.getLogger("replay")
        logger.debug("Calculating criterion")
        criterion_value = calculate_criterion_value(criterion, recs, split_data.test_dataset.interactions)
        logger.debug("%s=%.6f", criterion, criterion_value)
        return criterion_value

    def __call__(self, trial: "Trial") -> float:
        """
        Calculate criterion for ``optuna``.

        :param trial: current trial
        :return: criterion value
        """
        return self.objective_calculator(trial=trial, **self.kwargs)



import numpy as np
import pandas as pd

from replay.models.extensions.ann.index_inferers.base_inferer import IndexInferer
from replay.models.extensions.ann.index_inferers.utils import get_csr_matrix
from replay.models.extensions.ann.utils import create_nmslib_index_instance
from replay.utils import PYSPARK_AVAILABLE, PandasDataFrame, SparkDataFrame
from replay.utils.session_handler import State

if PYSPARK_AVAILABLE:
    from pyspark.sql.pandas.functions import pandas_udf


class NmslibFilterIndexInferer(IndexInferer):
    """Nmslib index inferer with filter seen items. Infers nmslib hnsw index."""

    def infer(self, vectors: SparkDataFrame, features_col: str, k: int) -> SparkDataFrame:  # noqa: ARG002
        _index_store = self.index_store
        index_params = self.index_params

        index_store_broadcast = State().session.sparkContext.broadcast(_index_store)

        @pandas_udf(self.udf_return_type)
        def infer_index_udf(
            user_idx: pd.Series,
            vector_items: pd.Series,
            vector_ratings: pd.Series,
            num_items: pd.Series,
            seen_item_ids: pd.Series,
        ) -> PandasDataFrame:  # pragma: no cover
            index_store = index_store_broadcast.value
            index = index_store.load_index(
                init_index=lambda: create_nmslib_index_instance(index_params),
                load_index=lambda index, path: index.loadIndex(path, load_data=True),
                configure_index=lambda index: (
                    index.setQueryTimeParams({"efSearch": index_params.ef_s}) if index_params.ef_s else None
                ),
            )

            # max number of items to retrieve per batch
            max_items_to_retrieve = num_items.max()

            user_vectors = get_csr_matrix(user_idx, vector_items, vector_ratings)

            neighbours = index.knnQueryBatch(
                user_vectors[user_idx.values, :],
                k=k + max_items_to_retrieve,
                num_threads=1,
            )

            neighbours_filtered = []
            for i, (item_idxs, distances) in enumerate(neighbours):
                non_seen_item_indexes = ~np.isin(item_idxs, seen_item_ids[i], assume_unique=True)
                neighbours_filtered.append(
                    (
                        (item_idxs[non_seen_item_indexes])[:k],
                        (distances[non_seen_item_indexes])[:k],
                    )
                )

            pd_res = PandasDataFrame(neighbours_filtered, columns=["item_idx", "distance"])

            """
            pd_res looks like
            item_idx       distances
            [1, 2, 3, ...] [-0.5, -0.3, -0.1, ...]
            [1, 3, 4, ...] [-0.1, -0.8, -0.2, ...]
            """

            return pd_res

        cols = [
            "user_idx",
            "vector_items",
            "vector_ratings",
            "num_items",
            "seen_item_idxs",
        ]

        res = vectors.select(
            "user_idx",
            infer_index_udf(*cols).alias("neighbours"),
        )

        res = self._unpack_infer_struct(res)

        return res



import functools
import inspect
import json
from pathlib import Path
from typing import Any, Callable, Union

from polars import from_pandas as pl_from_pandas

from replay.data.dataset import Dataset
from replay.preprocessing import (
    LabelEncoder,
    LabelEncodingRule,
)
from replay.splitters import (
    ColdUserRandomSplitter,
    KFolds,
    LastNSplitter,
    NewUsersSplitter,
    RandomSplitter,
    RatioSplitter,
    TimeSplitter,
    TwoStageSplitter,
)
from replay.utils import (
    TORCH_AVAILABLE,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)
from replay.utils.spark_utils import (
    convert2spark as pandas_to_spark,
    spark_to_pandas,
)

SavableObject = Union[
    ColdUserRandomSplitter,
    KFolds,
    LastNSplitter,
    NewUsersSplitter,
    RandomSplitter,
    RatioSplitter,
    TimeSplitter,
    TwoStageSplitter,
    Dataset,
    LabelEncoder,
    LabelEncodingRule,
]

if TORCH_AVAILABLE:
    from replay.data.nn import PandasSequentialDataset, PolarsSequentialDataset, SequenceTokenizer

    SavableObject = Union[
        SavableObject,
        SequenceTokenizer,
        PandasSequentialDataset,
        PolarsSequentialDataset,
    ]


def save_to_replay(obj: SavableObject, path: Union[str, Path]) -> None:
    """
    General function to save RePlay models, splitters and tokenizer.

    :param path: Path to save the object.
    """
    obj.save(path)


def load_from_replay(path: Union[str, Path], **kwargs) -> SavableObject:
    """
    General function to load RePlay models, splitters and tokenizer.

    :param path: Path to save the object.
    """
    path = Path(path).with_suffix(".replay").resolve()
    with open(path / "init_args.json") as file:
        class_name = json.loads(file.read())["_class_name"]
    obj_type = globals()[class_name]
    obj = obj_type.load(path, **kwargs)

    return obj


def _check_if_dataframe(var: Any):
    if not isinstance(var, (SparkDataFrame, PolarsDataFrame, PandasDataFrame)):
        msg = f"Object of type {type(var)} is not a dataframe of known type (can be pandas|spark|polars)"
        raise ValueError(msg)


def check_if_dataframe(*args_to_check: str) -> Callable[..., Any]:
    def decorator_func(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrap_func(*args: Any, **kwargs: Any) -> Any:
            extended_kwargs = {}
            extended_kwargs.update(kwargs)
            extended_kwargs.update(dict(zip(inspect.signature(func).parameters.keys(), args)))
            # add default param values to dict with arguments
            extended_kwargs.update(
                {
                    x.name: x.default
                    for x in inspect.signature(func).parameters.values()
                    if x.name not in extended_kwargs and x.default is not x.empty
                }
            )
            vals_to_check = [extended_kwargs[_arg] for _arg in args_to_check]
            for val in vals_to_check:
                _check_if_dataframe(val)
            return func(*args, **kwargs)

        return wrap_func

    return decorator_func


@check_if_dataframe("data")
def convert2pandas(
    data: Union[SparkDataFrame, PolarsDataFrame, PandasDataFrame], allow_collect_to_master: bool = False
) -> PandasDataFrame:
    """
    Convert the spark|polars DataFrame to a pandas.DataFrame.
    Returns unchanged dataframe if the input is already of type pandas.DataFrame.

    :param data: The dataframe to convert. Can be polars|spark|pandas DataFrame.
    :param allow_collect_to_master: If set to False (default) raises a warning
        about collecting parallelized data to the master node.
    """
    if isinstance(data, PandasDataFrame):
        return data

    if isinstance(data, PolarsDataFrame):
        return data.to_pandas()
    if isinstance(data, SparkDataFrame):
        return spark_to_pandas(data, allow_collect_to_master, from_constructor=False)


@check_if_dataframe("data")
def convert2polars(
    data: Union[SparkDataFrame, PolarsDataFrame, PandasDataFrame], allow_collect_to_master: bool = False
) -> PolarsDataFrame:
    """
    Convert the spark|pandas DataFrame to a polars.DataFrame.
    Returns unchanged dataframe if the input is already of type polars.DataFrame.

    :param data: The dataframe to convert. Can be spark|pandas|polars DataFrame.
    :param allow_collect_to_master: If set to False (default) raises a warning
        about collecting parallelized data to the master node.
    """
    if isinstance(data, PolarsDataFrame):
        return data

    if isinstance(data, PandasDataFrame):
        return pl_from_pandas(data)
    if isinstance(data, SparkDataFrame):
        return pl_from_pandas(spark_to_pandas(data, allow_collect_to_master, from_constructor=False))


@check_if_dataframe("data")
def convert2spark(data: Union[SparkDataFrame, PolarsDataFrame, PandasDataFrame]) -> SparkDataFrame:
    """
    Convert the pandas|polars DataFrame to a pysaprk.sql.DataFrame.
    Returns unchanged dataframe if the input is already of type pysaprk.sql.DataFrame.

    :param data: The dataframe to convert. Can be pandas|polars|spark Datarame.
    """
    if isinstance(data, (PandasDataFrame, SparkDataFrame)):
        return pandas_to_spark(data)
    if isinstance(data, PolarsDataFrame):
        return pandas_to_spark(data.to_pandas())



import logging
import os
import shutil
import weakref
from pathlib import Path
from typing import Any, Callable

from pyarrow import fs

from .base_index_store import IndexStore

logger = logging.getLogger("replay")


class SharedDiskIndexStore(IndexStore):
    """Class that responsible for index store in shared disk.
    It can also be used with a local disk when the driver and executors
    are running on the same machine."""

    def __init__(self, warehouse_dir: str, index_dir: str, cleanup: bool = True):
        self.index_dir_path = os.path.join(warehouse_dir, index_dir)
        super().__init__(cleanup)
        if self.cleanup:
            logger.debug(
                "Index directory %s is marked for deletion via weakref.finalize()",
                self.index_dir_path,
            )
            weakref.finalize(self, shutil.rmtree, self.index_dir_path)

    def load_index(
        self,
        init_index: Callable[[], None],
        load_index: Callable[[Any, str], None],
        configure_index: Callable[[Any], None],
    ):  # pragma: no cover
        if self._index:
            return self._index

        self._index = init_index()

        temp_file_path = os.path.join(self.index_dir_path, "index")
        load_index(self._index, temp_file_path)
        configure_index(self._index)

        return self._index

    def save_to_store(self, save_index: Callable[[str], None]):
        Path(self.index_dir_path).mkdir(parents=True, exist_ok=True)
        temp_file_path = os.path.join(self.index_dir_path, "index")
        save_index(temp_file_path)

    def dump_index(self, target_path: str):
        destination_filesystem, target_path = fs.FileSystem.from_uri(target_path)
        target_path = os.path.join(target_path, "index_files")
        destination_filesystem.create_dir(target_path)
        fs.copy_files(
            self.index_dir_path,
            target_path,
            source_filesystem=fs.LocalFileSystem(),
            destination_filesystem=destination_filesystem,
        )



from dataclasses import dataclass
from typing import Optional


@dataclass
class BaseHnswParam:
    """
    Base hnsw params.
    """

    space: str
    m: int = 200
    ef_c: int = 20000
    post: int = 0
    ef_s: Optional[int] = None

    def init_meta_as_dict(self) -> dict:
        """
        Returns meta-information for class instance initialization. Used to save the entity to disk.
        :return: dictionary with init meta.
        """
        return {
            "module": type(self).__module__,
            "class": type(self).__name__,
            "init_args": {
                "space": self.space,
                "m": self.m,
                "ef_c": self.ef_c,
                "post": self.post,
                "ef_s": self.ef_s,
            },
        }



from replay.utils import TORCH_AVAILABLE

if TORCH_AVAILABLE:
    from .schema import MutableTensorMap, TensorFeatureInfo, TensorFeatureSource, TensorMap, TensorSchema
    from .sequence_tokenizer import SequenceTokenizer
    from .sequential_dataset import PandasSequentialDataset, PolarsSequentialDataset, SequentialDataset
    from .torch_sequential_dataset import (
        DEFAULT_GROUND_TRUTH_PADDING_VALUE,
        DEFAULT_TRAIN_PADDING_VALUE,
        TorchSequentialBatch,
        TorchSequentialDataset,
        TorchSequentialValidationBatch,
        TorchSequentialValidationDataset,
    )

    __all__ = [
        "DEFAULT_GROUND_TRUTH_PADDING_VALUE",
        "DEFAULT_TRAIN_PADDING_VALUE",
        "MutableTensorMap",
        "PandasSequentialDataset",
        "PolarsSequentialDataset",
        "SequenceTokenizer",
        "SequentialDataset",
        "TensorFeatureInfo",
        "TensorFeatureSource",
        "TensorMap",
        "TensorSchema",
        "TorchSequentialBatch",
        "TorchSequentialDataset",
        "TorchSequentialValidationBatch",
        "TorchSequentialValidationDataset",
    ]



from typing import Optional, Union, cast

import numpy as np
import pandas as pd
import torch

from replay.data.nn import SequentialDataset

from ._base import BasePostProcessor


class RemoveSeenItems(BasePostProcessor):
    """
    Filters out the items that already have been seen in dataset.
    """

    def __init__(self, sequential: SequentialDataset) -> None:
        super().__init__()
        self._sequential = sequential
        self._apply_candidates = False
        self._candidates = None

    def on_validation(
        self, query_ids: torch.LongTensor, scores: torch.Tensor, ground_truth: torch.LongTensor
    ) -> tuple[torch.LongTensor, torch.Tensor, torch.LongTensor]:
        """
        Validation step.

        :param query_ids: query id sequence
        :param scores: calculated logits
        :param ground_truth: ground truth dataset

        :returns: modified query ids and scores and ground truth dataset
        """
        self._apply_candidates = False
        modified_scores = self._compute_scores(query_ids, scores)
        return query_ids, modified_scores, ground_truth

    def on_prediction(self, query_ids: torch.LongTensor, scores: torch.Tensor) -> tuple[torch.LongTensor, torch.Tensor]:
        """
        Prediction step.

        :param query_ids: query id sequence
        :param scores: calculated logits

        :returns: modified query ids and scores
        """
        self._apply_candidates = True
        modified_scores = self._compute_scores(query_ids, scores)
        return query_ids, modified_scores

    def _compute_scores(self, query_ids: torch.LongTensor, scores: torch.Tensor) -> torch.Tensor:
        flat_seen_item_ids = self._get_flat_seen_item_ids(query_ids)
        return self._fill_item_ids(scores.clone(), flat_seen_item_ids, -np.inf)

    def _fill_item_ids(
        self,
        scores: torch.Tensor,
        flat_item_ids: torch.LongTensor,
        value: float,
    ) -> torch.Tensor:
        flat_item_ids_on_device = flat_item_ids.to(scores.device)

        if self._apply_candidates and self._candidates is not None:
            item_count = self._sequential.schema.item_id_features.item().cardinality
            assert item_count
            _scores = torch.full((scores.shape[0], item_count), -float("inf")).to(scores.device)
            _scores[:, self._candidates] = torch.reshape(scores, _scores[:, self._candidates].shape)
            scores = _scores
        if scores.is_contiguous():
            scores.view(-1)[flat_item_ids_on_device] = value
        else:
            flat_scores = scores.flatten()
            flat_scores[flat_item_ids_on_device] = value
            scores = flat_scores.reshape(scores.shape)
        return scores

    def _get_flat_seen_item_ids(self, query_ids: torch.LongTensor) -> torch.LongTensor:
        query_ids_np = query_ids.flatten().cpu().numpy()

        item_count = self._sequential.schema.item_id_features.item().cardinality
        assert item_count
        item_id_feature_name = self._sequential.schema.item_id_feature_name
        assert item_id_feature_name

        item_id_sequences = self._sequential.get_sequence_by_query_id(query_ids_np, item_id_feature_name)

        for i in range(item_id_sequences.shape[0]):
            item_id_sequences[i] = item_id_sequences[i].copy() + i * item_count

        flat_seen_item_ids_np = np.concatenate(item_id_sequences)
        return torch.LongTensor(flat_seen_item_ids_np)

    @property
    def candidates(self) -> Union[torch.LongTensor, None]:
        """
        Returns tensor of item ids to calculate scores.
        """
        return self._candidates

    @candidates.setter
    def candidates(self, candidates: Optional[torch.LongTensor] = None) -> None:
        """
        Sets tensor of item ids to calculate scores.
        :param candidates: Tensor of item ids to calculate scores.
        """
        self._candidates = candidates


class SampleItems(BasePostProcessor):
    """
    Generates negative samples to compute sampled metrics
    """

    def __init__(
        self,
        grouped_validation_items: pd.DataFrame,
        user_col: str,
        item_col: str,
        items_list: np.ndarray,
        sample_count: int,
    ) -> None:
        self.items_set = set(items_list)
        self.sample_count = sample_count
        users = grouped_validation_items[user_col].to_numpy()
        items = grouped_validation_items[item_col].to_numpy()
        self.items_list: list[set[int]] = [set() for _ in range(users.shape[0])]
        for i in range(users.shape[0]):
            self.items_list[users[i]] = set(items[i])

    def on_validation(
        self, query_ids: torch.LongTensor, scores: torch.Tensor, ground_truth: torch.LongTensor
    ) -> tuple[torch.LongTensor, torch.Tensor, torch.LongTensor]:
        """
        Validation step.

        :param query_ids: query id sequence
        :param scores: calculated logits
        :param ground_truth: ground truth dataset

        :returns: modified query ids and scores and ground truth dataset
        """
        modified_score = self._compute_score(query_ids, scores, ground_truth)
        return query_ids, modified_score, ground_truth

    def on_prediction(self, query_ids: torch.LongTensor, scores: torch.Tensor) -> tuple[torch.LongTensor, torch.Tensor]:
        """
        Prediction step.

        :param query_ids: query id sequence
        :param scores: calculated logits

        :returns: modified query ids and scores
        """
        modified_score = self._compute_score(query_ids, scores, None)
        return query_ids, modified_score

    def _compute_score(
        self, query_ids: torch.LongTensor, scores: torch.Tensor, ground_truth: Optional[torch.LongTensor]
    ) -> torch.Tensor:
        batch_size = query_ids.shape[0]
        item_ids = ground_truth.cpu().numpy() if ground_truth is not None else None
        candidate_ids: list[torch.Tensor] = []
        candidate_labels: list[torch.Tensor] = []
        for user in range(batch_size):
            ground_truth_items = set(item_ids[user]) if ground_truth is not None else set()
            sample, label = self._generate_samples_for_user(ground_truth_items, self.items_list[user])
            candidate_ids.append(sample)
            candidate_labels.append(label)

        candidate_ids_torch = cast(torch.LongTensor, torch.stack(candidate_ids).to(scores.device).long())

        return self._fill_scores(candidate_ids_torch, scores, np.inf)

    def _fill_scores(self, ids: torch.LongTensor, scores: torch.Tensor, value: float) -> torch.Tensor:
        row_count = ids.shape[0]
        item_count = scores.shape[1]
        stride = torch.arange(0, row_count * item_count, item_count).to(scores.device).reshape(-1, 1)
        strided_ids = ids + stride
        new_scores = scores.flatten()
        new_scores[strided_ids.flatten()] = value

        return new_scores.reshape_as(scores)

    def _generate_samples_for_user(
        self, ground_truth_items: set[int], input_items: set[int]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        negative_sample_count = self.sample_count - len(ground_truth_items)
        assert negative_sample_count > 0

        # Negative samples in the original paper excludes the input sequences.
        random_items = np.random.choice(list(self.items_set - input_items - ground_truth_items), negative_sample_count)
        samples_list = list(random_items) + list(ground_truth_items)

        samples = torch.tensor(samples_list, dtype=torch.long)
        labels = torch.cat(
            [
                torch.zeros(negative_sample_count, dtype=torch.bool),
                torch.ones(len(ground_truth_items), dtype=torch.bool),
            ]
        )

        return (
            samples,
            labels,
        )



"""
Hyperparameter optimization of models
"""

from replay.utils.types import OPTUNA_AVAILABLE

from .optuna_mixin import IsOptimizible

if OPTUNA_AVAILABLE:
    from .optuna_objective import ItemKNNObjective, ObjectiveWrapper

    __all__ = ["IsOptimizible", "ItemKNNObjective", "ObjectiveWrapper"]
else:
    __all__ = ["IsOptimizible"]



from .types import (
    ANN_AVAILABLE,
    LIGHTFM_AVAILABLE,
    OPENVINO_AVAILABLE,
    OPTUNA_AVAILABLE,
    PYSPARK_AVAILABLE,
    TORCH_AVAILABLE,
    DataFrameLike,
    FeatureUnavailableError,
    FeatureUnavailableWarning,
    IntOrList,
    MissingImport,
    NumType,
    PandasDataFrame,
    PolarsDataFrame,
    SparkDataFrame,
)



from .base_metric import Metric


class MRR(Metric):
    """
    Mean Reciprocal Rank -- Reciprocal Rank is the inverse position of the
    first relevant item among top-k recommendations,
    :math:`\\frac{1}{rank_i}`. This value is aggregated by all users.

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> groundtruth
       query_id  item_id
    0         1        5
    1         1        6
    2         1        7
    3         1        8
    4         1        9
    5         1       10
    6         2        6
    7         2        7
    8         2        4
    9         2       10
    10        2       11
    11        3        1
    12        3        2
    13        3        3
    14        3        4
    15        3        5
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> MRR(2)(recommendations, groundtruth)
    {'MRR@2': 0.5}
    >>> MRR(2, mode=PerUser())(recommendations, groundtruth)
    {'MRR-PerUser@2': {1: 0.5, 2: 0.0, 3: 1.0}}
    >>> MRR(2, mode=Median())(recommendations, groundtruth)
    {'MRR-Median@2': 0.5}
    >>> MRR(2, mode=ConfidenceInterval(alpha=0.95))(recommendations, groundtruth)
    {'MRR-ConfidenceInterval@2': 0.565792867038086}
    <BLANKLINE>
    """

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], ground_truth: list, pred: list) -> list[float]:
        if not ground_truth or not pred:
            return [0.0 for _ in ks]
        set_gt = set(ground_truth)
        res = []
        for k in ks:
            ans = 0.0
            for index, item_id in enumerate(pred[:k]):
                if item_id in set_gt:
                    ans = 1 / (index + 1)
                    break
            res.append(ans)
        return res



from typing import Optional

from replay.data import Dataset
from replay.utils import PYSPARK_AVAILABLE

from .base_rec import NonPersonalizedRecommender

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf


class RandomRec(NonPersonalizedRecommender):
    """
    Recommend random items, either weighted by item popularity or uniform.

    .. math::
        P\\left(i\\right)\\propto N_i + \\alpha

    :math:`N_i` --- number of users who rated item :math:`i`

    :math:`\\alpha` --- bigger :math:`\\alpha` values increase amount of rare items in recommendations.
        Must be bigger than -1. Default value is :math:`\\alpha = 0`.

    Model without seed provides non-determenistic recommendations,
    model with fixed seed provides reproducible recommendataions.

    As the recommendations from `predict` are cached, save them to disk, or create a checkpoint
    and unpersist them to get different recommendations after another `predict` call.

    >>> from replay.utils.session_handler import get_spark_session, State
    >>> spark = get_spark_session(1, 1)
    >>> state = State(spark)

    >>> import pandas as pd
    >>> from replay.data.dataset import Dataset, FeatureSchema, FeatureInfo, FeatureHint, FeatureType
    >>> from replay.utils.spark_utils import convert2spark
    >>>
    >>> interactions = convert2spark(pd.DataFrame({
    ...     "user_id": [1, 1, 2, 2, 3, 4],
    ...     "item_id": [1, 2, 2, 3, 3, 3]
    ... }))
    >>> interactions.show()
    +-------+-------+
    |user_id|item_id|
    +-------+-------+
    |      1|      1|
    |      1|      2|
    |      2|      2|
    |      2|      3|
    |      3|      3|
    |      4|      3|
    +-------+-------+
    <BLANKLINE>
    >>> random_pop = RandomRec(distribution="popular_based", alpha=-1)
    Traceback (most recent call last):
     ...
    ValueError: alpha must be bigger than -1

    >>> random_pop = RandomRec(distribution="abracadabra")
    Traceback (most recent call last):
     ...
    ValueError: distribution can be one of [popular_based, relevance, uniform]

    >>> feature_schema = FeatureSchema(
    ...     [
    ...         FeatureInfo(
    ...             column="user_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.QUERY_ID,
    ...         ),
    ...         FeatureInfo(
    ...             column="item_id",
    ...             feature_type=FeatureType.CATEGORICAL,
    ...             feature_hint=FeatureHint.ITEM_ID,
    ...         ),
    ...     ]
    ... )
    >>> dataset = Dataset(feature_schema, interactions)
    >>> random_pop = RandomRec(distribution="popular_based", alpha=1.0, seed=777)
    >>> random_pop.fit(dataset)
    >>> random_pop.item_popularity.show()
    +-------+------------------+
    |item_id|            rating|
    +-------+------------------+
    |      1|0.2222222222222222|
    |      2|0.3333333333333333|
    |      3|0.4444444444444444|
    +-------+------------------+
    <BLANKLINE>
    >>> recs = random_pop.predict(dataset, 2)
    >>> recs.show()
    +-------+-------+-------------------+
    |user_id|item_id|             rating|
    +-------+-------+-------------------+
    |      1|      3| 0.5403400662326382|
    |      2|      1| 0.8999281214873164|
    |      3|      1|  0.909252584485465|
    |      3|      2|0.14871615041196307|
    |      4|      2| 0.9923148665602072|
    |      4|      1| 0.4814574953690183|
    +-------+-------+-------------------+
    <BLANKLINE>
    >>> recs = random_pop.predict(dataset, 2, queries=[1], items=[7, 8])
    >>> recs.show()
    +-------+-------+-------------------+
    |user_id|item_id|             rating|
    +-------+-------+-------------------+
    |      1|      7|0.25547770543750425|
    |      1|      8|0.12755340075617772|
    +-------+-------+-------------------+
    <BLANKLINE>
    >>> random_pop = RandomRec(seed=555)
    >>> random_pop.fit(dataset)
    >>> random_pop.item_popularity.show()
    +-------+------------------+
    |item_id|            rating|
    +-------+------------------+
    |      1|0.3333333333333333|
    |      2|0.3333333333333333|
    |      3|0.3333333333333333|
    +-------+------------------+
    <BLANKLINE>
    """

    _search_space = {
        "distribution": {
            "type": "categorical",
            "args": ["popular_based", "relevance", "uniform"],
        },
        "alpha": {"type": "uniform", "args": [-0.5, 100]},
    }
    sample: bool = True

    def __init__(
        self,
        distribution: str = "uniform",
        alpha: float = 0.0,
        seed: Optional[int] = None,
        add_cold_items: bool = True,
        cold_weight: float = 0.5,
    ):
        """
        :param distribution: recommendation strategy:
            "uniform" - all items are sampled uniformly
            "popular_based" - recommend popular items more
        :param alpha: bigger values adjust model towards less popular items
        :param seed: random seed
        :param add_cold_items: flag to consider cold items in recommendations building
            if present in `items` parameter of `predict` method
            or `pairs` parameter of `predict_pairs` methods.
            If true, cold items are assigned relevance equals to the less relevant item relevance
            multiplied by `cold_weight` and may appear among top-K recommendations.
            Otherwise cold items are filtered out.
            Could be changed after model training by setting the `add_cold_items` attribute.
        :param cold_weight: if `add_cold_items` is True,
            cold items are added with reduced relevance.
            The relevance for cold items is equal to the relevance
            of a least relevant item multiplied by a `cold_weight` value.
            `Cold_weight` value should be in interval (0, 1].
        """
        if distribution not in ("popular_based", "relevance", "uniform"):
            msg = "distribution can be one of [popular_based, relevance, uniform]"
            raise ValueError(msg)
        if alpha <= -1.0 and distribution == "popular_based":
            msg = "alpha must be bigger than -1"
            raise ValueError(msg)
        self.distribution = distribution
        self.alpha = alpha
        self.seed = seed
        super().__init__(add_cold_items=add_cold_items, cold_weight=cold_weight)

    @property
    def _init_args(self):
        return {
            "distribution": self.distribution,
            "alpha": self.alpha,
            "seed": self.seed,
            "add_cold_items": self.add_cold_items,
            "cold_weight": self.cold_weight,
        }

    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        if self.rating_column is None:
            self.rating_column = "rating"
        if self.distribution == "popular_based":
            self.item_popularity = (
                dataset.interactions.groupBy(self.item_column)
                .agg(sf.countDistinct(self.query_column).alias("user_count"))
                .select(
                    sf.col(self.item_column),
                    (sf.col("user_count").astype("float") + sf.lit(self.alpha)).alias(self.rating_column),
                )
            )
        elif self.distribution == "relevance":
            self.item_popularity = (
                dataset.interactions.groupBy(self.item_column)
                .agg(sf.sum(self.rating_column).alias(self.rating_column))
                .select(self.item_column, self.rating_column)
            )
        else:
            self.item_popularity = (
                dataset.interactions.select(self.item_column).distinct().withColumn(self.rating_column, sf.lit(1.0))
            )
        self.item_popularity = self.item_popularity.withColumn(
            self.rating_column,
            sf.col(self.rating_column) / self.item_popularity.agg(sf.sum(self.rating_column)).first()[0],
        )
        self.item_popularity.cache().count()
        self.fill = self._calc_fill(self.item_popularity, self.cold_weight, self.rating_column)
